from indica.configs.readers.machineconf import MachineConf


class ST40Conf(MachineConf):
    def __init__(self):
        self.MACHINE_DIMS = ((0.15, 0.85), (-0.75, 0.75))
        self.INSTRUMENT_METHODS = {
            "efit": "get_equilibrium",
            "xrcs": "get_helike_spectroscopy",
            "pi": "get_spectrometer",
            "tws_c": "get_spectrometer",
            "cxff_pi": "get_charge_exchange",
            "cxff_tws_c": "get_charge_exchange",
            "cxff_tws_b": "get_charge_exchange",
            "cxqf_tws_c": "get_charge_exchange",
            "lines": "get_diode_filters",
            "smmh": "get_interferometry",
            "nirh": "get_interferometry",
            "ts": "get_thomson_scattering",
            "sxr_spd": "get_radiation",
            "sxrc_xy1": "get_radiation",
            "sxrc_xy2": "get_radiation",
            "sxrc_rz1": "get_radiation",
            "sxrc_rz2": "get_radiation",
            "blom_xy1": "get_radiation",
            "blom_rz1": "get_radiation",
            "blom_dv1": "get_radiation",
            "astra": "get_astra",
            "ppts": "get_profile_fits",
            "zeff_brems": "get_zeff",
        }
        self.QUANTITIES_PATH = {
            "get_equilibrium": {
                "t": ":time",
                "psin": ".profiles.psi_norm:xpsn",
                "R": ".psi2d:rgrid",
                "z": ".psi2d:zgrid",
                "f": ".profiles.psi_norm:f",
                "psi_axis": ".global:faxs",
                "psi_boundary": ".global:fbnd",
                "ftor": ".profiles.psi_norm:ftor",
                "rmji": ".profiles.psi_norm:rmji",
                "rmjo": ".profiles.psi_norm:rmjo",
                "psi": ".psi2d:psi",
                "vjac": ".profiles.psi_norm:vjac",
                "ajac": ".profiles.psi_norm:ajac",
                "rmag": ".global:rmag",
                "rgeo": ".global:rgeo",
                "rbnd": ".p_boundary:rbnd",
                "zmag": ".global:zmag",
                "zbnd": ".p_boundary:zbnd",
                "ipla": ".constraints.ip:cvalue",
                "wp": ".virial:wp",
                "df": ".constraints.df:cvalue",
            },
            "get_helike_spectroscopy": {
                "t": ":time",
                "wavelength": ":wavelen",
                "location": ".geometry:location",
                "direction": ".geometry:direction",
                "ti_w": ".global:ti_w",
                "ti_z": ".global:ti_z",
                "te_n3w": ".global:te_n3w",
                "te_kw": ".global:te_kw",
                "spectra_raw": ":intens",
                "spectra": ":spec_rad",
                "int_w": ".global:int_w",
                "int_k": ".global:int_k",
                "int_tot": ".global:int_tot",
                "int_n3": ".global:int_n3",
                "background": ".global:back_avg",
            },
            "get_interferometry": {
                "t": ":time",
                "location": ".geometry:location",
                "direction": ".geometry:direction",
                "ne": ".global:ne_int",
            },
            "get_diode_filters": {
                "t": ":time",
                "label": ":label",
                "location": ".geometry:location",
                "direction": ".geometry:direction",
                "brightness": ":data",
            },
            "get_radiation": {
                "t": ":time",
                "location": ".geometry:location",
                "direction": ".geometry:direction",
                "brightness": ".profiles:emission",
            },
            "get_charge_exchange": {
                "t": ":time",
                "wavelength": ":wavelen",
                "x": ":x",
                "y": ":y",
                "z": ":z",
                "R": ":R",
                "int": ".profiles:int",
                "ti": ".profiles:ti",
                "vtor": ".profiles:vtor",
                "spectra": ":spectra",
                "fit": ":full_fit",
            },
            "get_spectrometer": {
                "t": ":time",
                "location": ".geometry:location",
                "direction": ".geometry:direction",
                "wavelength": ":wavelen",
                "spectra": ":emission",
            },
            "get_thomson_scattering": {
                "t": ":time",
                "x": ":x",
                "y": ":y",
                "z": ":z",
                "R": ":R",
                "ne": ".profiles:ne",
                "te": ".profiles:te",
                "pe": ".profiles:pe",
                "chi2": ".profiles:chi2",
            },
            "get_profile_fits": {
                "t": ":time",
                "rhop": ".profiles.psi_norm:rhop",
                "R": ".profiles.r_midplane:rpos",
                "z": ".profiles.r_midplane:zpos",
                "rhop_data": ".profiles.inputs:rhop",
                "R_data": ".profiles.inputs:rpos",
                "z_data": ".profiles.inputs:zpos",
                "ne_rhop": ".profiles.psi_norm:ne",
                "te_rhop": ".profiles.psi_norm:te",
                "pe_rhop": ".profiles.psi_norm:pe",
                "ne_R": ".profiles.r_midplane:ne",
                "te_R": ".profiles.r_midplane:te",
                "pe_R": ".profiles.r_midplane:pe",
                "ne_data": ".profiles.inputs:ne",
                "te_data": ".profiles.inputs:te",
                "pe_data": ".profiles.inputs:pe",
                "R_shift": ".global:rshift",
            },
            "get_zeff": {
                "t": ":time",
                "rhop": ".profiles.psi_norm:rhop",
                "R_shift": ".global:rshift",
                "zeff_avrg": ".global:zeff",
                "zeff_hi": ".global:zeff_hi",
                "zeff_low": ".global:zeff_low",
                "zeff": ".profiles.psi_norm:zeff",
            },
            "get_astra": {
                "psi_axis": ".global:faxs",
                "psi_boundary": ".global:fbnd",
                "rmag": ".global:rmag",
                "rgeo": ".global:rgeo",
                "zmag": ".global:zmag",
                "zgeo": ".global:zgeo",
                "wp": ".global:wth",
                "ipla": ".global:ipl",
                "upl": ".global:upl",
                "wth": ".global:wth",
                "wtherm": ".global:wtherm",
                "wfast": ".global:wfast",
                "df": ".global.df",
                "pnb": ".global:pnb",  # W
                "pabs": ".global:pabs",  # W
                "p_oh": ".global:p_oh",  # W
                "q": ".profiles.psi_norm:q",
                "f": ".profiles.psi_norm:fpol",
                "ftor": ".profiles.psi_norm:ftor",
                "psi_1d": ".profiles.psi_norm:psi",
                "p": ".profiles.psi_norm:p",
                "volume": ".profiles.psi_norm:volume",
                "area": ".profiles.psi_norm:areat",
                "sigmapar": ".profiles.psi_norm:sigmapar",  # 1/(Ohm*m)
                "psi": ".psi2d:psi",
                "rbnd": ".p_boundary:rbnd",
                "zbnd": ".p_boundary:zbnd",
                "elon": ".profiles.astra:elon",  #
                "j_bs": ".profiles.astra:j_bs",  # MA/m2
                "j_nbi": ".profiles.astra:j_nbi",  # MA/m2
                "j_oh": ".profiles.astra:j_oh",  # MA/m2
                "j_rf": ".profiles.astra:j_rf",  # MA/m2
                "j_tot": ".profiles.astra:j_tot",  # MA/m2
                "ne": ".profiles.astra:ne",  # 10^19 m^-3
                "ni": ".profiles.astra:ni",  # 10^19 m^-3
                "nf": ".profiles.astra:nf",  # 10^19 m^-3
                "n_d": ".profiles.astra:n_d",  # 10E19/m3
                "n_t": ".profiles.astra:n_t",  # 10E19/m3
                "omega_tor": ".profiles.astra:omega_tor",  # 1/s
                "qe": ".profiles.astra:qe",  # MW
                "qi": ".profiles.astra:qi",  # MW
                "qn": ".profiles.astra:qn",  # 10^19/s
                "qnbe": ".profiles.astra:qnbe",  # MW/m3
                "qnbi": ".profiles.astra:qnbi",  # MW/m3
                "q_oh": ".profiles.astra:q_oh",  # MW/m3
                "q_rf": ".profiles.astra:q_rf",  # MW/m3
                "rhot": ".profiles.astra:rho",  # ASTRA rho-toroidal
                "rmid": ".profiles.astra:rmid",  # Centre of flux surfaces, m
                "rminor": ".profiles.astra:rminor",  # minor radius, m
                "sbm": ".profiles.astra:sbm",  # 10^19/m^3/s
                "spel": ".profiles.astra:spel",  # 10^19/m^3/s
                "stot": ".profiles.astra:stot",  # 10^19/s/m3
                "swall": ".profiles.astra:swall",  # 10^19/m^3/s
                "te": ".profiles.astra:te",  # keV
                "ti": ".profiles.astra:ti",  # keV
                "tri": ".profiles.astra:tri",  # Triangularity (up/down symmetrized)
                "t_d": ".profiles.astra:t_d",  # keV
                "t_t": ".profiles.astra:t_t",  # keV
                "zeff": ".profiles.astra:zeff",
                "pblon": ".profiles.astra:pblon",
                "pbper": ".profiles.astra:pbper",
                "nn": ".profiles.astra:nn",  # 10^19/m^3
                "niz1": ".profiles.astra:niz1",  # 10^19/m^3
                "niz2": ".profiles.astra:niz2",  # 10^19/m^3
                "niz3": ".profiles.astra:niz3",  # 10^19/m^3
            },
        }



from typing import Tuple

import matplotlib.pylab as plt
import numpy as np
from scipy.interpolate import CubicSpline
from scipy.optimize import least_squares
import xarray as xr
from xarray import DataArray

from indica.converters import LineOfSightTransform
from indica.numpy_typing import ArrayLike
from indica.numpy_typing import LabeledArray
from indica.operators.centrifugal_asymmetry import centrifugal_asymmetry_2d_map

DataArrayCoords = Tuple[DataArray, DataArray]


class InvertPoloidalAsymmetry:
    def __init__(
        self,
        knots: LabeledArray = [0, 0.2, 0.5, 0.9, 1.0],
        xspl: ArrayLike = np.linspace(0, 1.05, 51),
        bounds_profile: tuple = (0, np.inf),
        bounds_asymmetry: tuple = (-np.inf, np.inf),
        bc_profile: str = "clamped",
        bc_asymmetry: str = "natural",
        default_perc_err: float = 0.05,
    ):

        # xknots must include 0, 1, and rho_max
        rho_max = np.max(xspl)
        _xknots = np.append(np.array(knots), [0.0, 1.0, rho_max])
        _xknots = np.unique(np.sort(_xknots))
        xknots = _xknots[np.where((_xknots >= -0) * (_xknots <= rho_max))[0]]

        # Profile knots to scan <= 1.
        mask_profile = (xknots >= 0) * (xknots <= 1.0)
        indx_profile = np.where(mask_profile)[0]
        nknots_profile = len(indx_profile)

        # Asymmetry knots to scan >0 and <1.
        mask_asymmetry = (xknots > 0) * (xknots < 1.0)
        indx_asymmetry = np.where(mask_asymmetry)[0] + indx_profile[-1]
        nknots_asymmetry = len(indx_asymmetry)

        # Define optimisation bounds
        lbounds = np.append(
            [bounds_profile[0]] * nknots_profile,
            [bounds_asymmetry[0]] * nknots_asymmetry,
        )
        ubounds = np.append(
            [bounds_profile[1]] * nknots_profile,
            [bounds_asymmetry[1]] * nknots_asymmetry,
        )
        bounds = (lbounds, ubounds)

        self.xspl = xspl
        self.xknots = xknots
        self.nknots = np.size(xknots)
        self.bounds = bounds
        self.mask_profile = mask_profile
        self.mask_asymmetry = mask_asymmetry
        self.indx_profile = indx_profile
        self.indx_asymmetry = indx_asymmetry
        self.bc_profile = bc_profile
        self.bc_asymmetry = bc_asymmetry
        self.default_perc_err = default_perc_err

    def __call__(
        self,
        los_integral: DataArray,
        los_transform: LineOfSightTransform,
        t: LabeledArray = None,
        debug: bool = False,
    ):
        """
        Estimates the poloidal distribution from line-of-sight integrals
        assuming the local quantity is poloidally asymmetric following
        Wesson's formula.

        Parameters
        ----------
        los_integral
            Measured LOS-integral from e.g. a pinhole camera.
        los_transform
            Line of sight transform for the forward model (must already have
            assigned equilibrium)
        knots
            The spline knots to use when fitting the emissivity data.
        """

        coords = [("rhop", self.xspl)]
        mask_profile, mask_asymmetry = self.mask_profile, self.mask_asymmetry
        indx_profile, indx_asymmetry = self.indx_profile, self.indx_asymmetry
        xknots, xspl = self.xknots, self.xspl
        equilibrium = los_transform.equilibrium
        bc_profile, bc_asymmetry = self.bc_profile, self.bc_asymmetry

        def evaluate(xknots, yconcat):
            """
            1. Separate profile and asymmetry knots + add boundaries
                asymmetry(rho>=1)
            2. Initialize and call splines.
            3. Map profile and asymmetry to 2D
            """
            yprofile = yconcat[indx_profile]
            yprofile = np.append(yprofile, 0.0)
            yasymmetry = yconcat[indx_asymmetry]
            yasymmetry = np.append(yasymmetry[0], yasymmetry)
            yasymmetry = np.append(yasymmetry, [yasymmetry[-1], yasymmetry[-1]])
            profile_spline = CubicSpline(
                xknots,
                yprofile,
                0,
                bc_profile,
            )
            profile_to_map = DataArray(profile_spline(xspl), coords=coords)
            profile_to_map = xr.where(profile_to_map >= 0, profile_to_map, 0.0)
            asymmetry_spline = CubicSpline(
                xknots,
                yasymmetry,
                0,
                bc_asymmetry,
            )
            asymmetry_parameter = DataArray(asymmetry_spline(xspl), coords=coords)
            profile_2d = centrifugal_asymmetry_2d_map(
                profile_to_map,
                asymmetry_parameter,
                equilibrium=equilibrium,
                t=t,
            )
            return profile_2d, profile_to_map, asymmetry_parameter

        def residuals(yknots_concat):
            """
            Calculate the residuals to minimize
            """
            profile_2d, _, _ = evaluate(xknots, yknots_concat)
            _bckc = los_transform.integrate_on_los(profile_2d, t=t)
            return (_data - _bckc) / _error

        if t is None:
            t = los_integral.t.values
        else:
            t = los_integral.t.interp(t=t).values

        if hasattr(los_integral, "error"):
            error = los_integral.error
        else:
            error = los_integral * self.default_perc_err

        # Make sure error is > 0 to avoid residuals -> inf
        error = xr.where(error > 0, error, np.nan)
        error = xr.where(np.isfinite(error), error, error.min())

        self.t = np.array(t, ndmin=1)
        self.los_transform = los_transform
        self.data = los_integral
        self.error = error

        # Normalise to have similar ranges for profile and asymmetry
        norm_factor = los_integral.max().values
        data = los_integral / norm_factor
        error = error / norm_factor

        # Initial guesses
        _guess_profile = data.sel(t=self.t[0]).mean().values
        _guess_asymmetry = 0.0
        guess_profile = np.full_like(xknots, _guess_profile)
        guess_asymmetry = np.full_like(xknots, _guess_asymmetry)
        yconcat = np.append(
            guess_profile[mask_profile],
            guess_asymmetry[mask_asymmetry],
        )

        profile = []
        asymmetry = []
        profile_2d = []
        bckc = []
        for t in self.t:
            _data = data.sel(t=t).values
            _error = error.sel(t=t).values
            _x = np.arange(len(_data))

            if debug:
                plt.ioff()
                plt.figure()
                plt.errorbar(_x, _data, _error, marker="o")

            fit = least_squares(
                residuals,
                yconcat,
                bounds=self.bounds,
                verbose=debug,
            )
            yconcat = fit.x

            _profile_2d, _profile_to_map, _asymmetry_parameter = evaluate(
                xknots, yconcat
            )
            _bckc = los_transform.integrate_on_los(_profile_2d, t=t)
            if debug:
                plt.plot(_x, _bckc, linewidth=2)
                plt.show()

            profile.append(_profile_to_map)
            asymmetry.append(_asymmetry_parameter)
            profile_2d.append(_profile_2d)
            bckc.append(_bckc)

        profile = xr.concat(profile, "t") * norm_factor
        asymmetry = xr.concat(asymmetry, "t")
        profile_2d = xr.concat(profile_2d, "t") * norm_factor
        bckc = xr.concat(bckc, "t") * norm_factor

        return profile_2d, bckc, profile, asymmetry



import numpy as np
from scipy.optimize import least_squares

from indica import Plasma


def match_interferometer_los_int(
    models: dict,
    plasma: Plasma,
    data: dict,
    t: float,
    optimise_for: dict = {"smmh1": ["ne"]},
    guess: float = 5.0e19,
    bounds: tuple = (1.0e17, 1.0e21),
    bckc: dict = {},
):
    def residuals(guess):
        plasma.Ne_prof.y0 = guess
        plasma.assign_profiles("electron_density", t=t)

        for instrument in optimise_for.keys():
            if instrument in models.keys():
                bckc[instrument] = models[instrument](t=t)

        resid = []
        for instrument in optimise_for.keys():
            if instrument in models.keys():
                for quantity in optimise_for[instrument]:
                    resid.append(
                        data[instrument][quantity].sel(t=t) - bckc[instrument][quantity]
                    )

        return (np.array(resid) ** 2).sum()

    least_squares(residuals, guess, bounds=bounds, method="dogbox")

    return plasma, bckc


def match_helike_spectroscopy_line_ratios(
    models: dict,
    plasma: Plasma,
    data: dict,
    t: float,
    optimise_for: dict = {"xrcs": ["int_k/int_w"]},
    guess: float = 1.0e3,
    bounds: tuple = (1.0e2, 10e4),
    bckc: dict = {},
):
    def residuals(guess):
        plasma.Te_prof.y0 = guess
        plasma.assign_profiles("electron_temperature", t=t)
        plasma.Ti_prof.y0 = guess
        plasma.assign_profiles("ion_temperature", t=t)

        for instrument in optimise_for.keys():
            if instrument in models.keys():
                bckc[instrument] = models[instrument](t=t)

        resid = []
        for instrument in optimise_for.keys():
            if instrument in models.keys():
                for quantity in optimise_for[instrument]:
                    resid.append(
                        data[instrument][quantity].sel(t=t)
                        - bckc[instrument][quantity].sel(t=t)
                    )

        return (np.array(resid) ** 2).sum()

    least_squares(residuals, guess, bounds=bounds, method="dogbox")

    return plasma, bckc


def match_helike_spectroscopy_ion_temperature(
    models: dict,
    plasma: Plasma,
    data: dict,
    t: float,
    optimise_for: dict = {"xrcs": ["ti_w"]},
    guess: float = 1.0e3,
    bounds: tuple = (1.0e2, 1.5e3),
    bckc: dict = {},
):
    def residuals(guess):
        plasma.Ti_prof.y0 = guess
        plasma.assign_profiles("ion_temperature", t=t)

        for instrument in optimise_for.keys():
            if instrument in models.keys():
                bckc[instrument] = models[instrument](t=t)

        resid = []
        for instrument in optimise_for.keys():
            if instrument in models.keys():
                for quantity in optimise_for[instrument]:
                    resid.append(
                        data[instrument][quantity].sel(t=t) - bckc[instrument][quantity]
                    )

        return (np.array(resid) ** 2).sum()

    least_squares(residuals, guess, bounds=bounds, method="dogbox")

    return plasma, bckc


def match_helike_spectroscopy_intensity(
    models: dict,
    plasma: Plasma,
    data: dict,
    t: float,
    optimise_for: dict = {"xrcs": ["int_w"]},
    guess: float = 1.0e15,
    bounds: tuple = (1.0e13, 1.0e19),
    bckc: dict = {},
    element: tuple = ("ar",),
):
    def residuals(guess):
        plasma.Nimp_prof.y0 = guess
        plasma.assign_profiles("impurity_density", t=t, element=element)

        for instrument in optimise_for.keys():
            if instrument in models.keys():
                bckc[instrument] = models[instrument](t=t)

        resid = []
        for instrument in optimise_for.keys():
            if instrument in models.keys():
                for quantity in optimise_for[instrument]:
                    resid.append(
                        data[instrument][quantity].sel(t=t) - bckc[instrument][quantity]
                    )

        return (np.array(resid) ** 2).sum()

    least_squares(residuals, guess, bounds=bounds, method="dogbox")

    return plasma, bckc



from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.gaussian_process import kernels
import xarray as xr
from xarray import DataArray

from indica.readers.read_st40 import ReadST40
from indica.utilities import FIG_PATH
from indica.utilities import set_axis_sci
from indica.utilities import set_plot_colors
from indica.utilities import set_plot_rcparams

CMAP, COLORS = set_plot_colors()


def gpr_fit(
    x: np.array,
    y: np.array,
    y_err: np.array,
    x_fit: np.array,
    kernel: kernels.Kernel,
):
    """
    Run GPR fit given input data and desired x-grid

    Parameters
    ----------
    x
        x-axis of the data
    y
        data value
    y_err
        data error
    x_fit
        x-axis for fitting
    kernel
        Kernel used

    Returns
    -------
    Fit and error on desired x-grid

    """

    isort = np.argsort(x)
    _x = np.sort(x)
    _y = np.interp(_x, x[isort], y[isort])
    _y_err = np.interp(_x, x[isort], y_err[isort])

    idx = np.isfinite(_y)

    _x = _x[idx].reshape(-1, 1)
    _y_err = _y_err[idx]
    _y = _y[idx].reshape(-1, 1)

    gpr = GaussianProcessRegressor(
        kernel=kernel,
        alpha=_y_err**2,
    )  # alpha is sigma^2
    gpr.fit(
        _x,
        _y,
    )

    _x_fit = x_fit.reshape(-1, 1)
    y_fit, y_fit_err = gpr.predict(_x_fit, return_std=True)

    return y_fit, y_fit_err, gpr


def plot_gpr_fit(
    data: DataArray,
    y_fit: DataArray,
    y_fit_err: DataArray,
    fig_style: str = "profiles",
    ylabel: str = "",
    xlabel: str = "",
    label: str = "",
    title: str = "",
    fig_name: str = "",
    save_fig: bool = False,
    color=None,
):
    set_plot_rcparams(fig_style)

    plt.figure()

    x_data = getattr(data, data.dims[0])
    x_fit = getattr(y_fit, data.dims[0])
    y_err = data.error

    plt.fill_between(
        x_fit,
        y_fit - y_fit_err,
        y_fit + y_fit_err,
        alpha=0.5,
        color=color,
    )
    plt.plot(x_fit, y_fit, color=color)
    plt.errorbar(x_data, data, y_err, linestyle="", color=color)
    plt.plot(x_data, data, marker="o", linestyle="", color=color, label=label)
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    plt.title(title)
    set_axis_sci()
    if len(label) > 0:
        plt.legend()

    if save_fig:
        plt.savefig(
            FIG_PATH + f"{fig_name}_GPR_t:{data.t.values:.3f}.png", bbox_inches="tight"
        )


def post_process_ts(binned_data, equilibrium, quant, pulse, split=""):
    rmag = binned_data["efit"]["rmag"]
    data = binned_data["ts"][quant]
    data["pulse"] = pulse
    data["quantity"] = quant

    data.transform.set_equilibrium(equilibrium)
    data.transform.convert_to_rho_theta(t=data.t)
    data["rho"] = data.transform.rho

    # Normalising
    if quant == "ne":
        data.values = data.values * 1e-19
        data["error"] = data.error * 1e-19
        data["unit"] = "n19"
    elif quant == "te":
        data.values = data.values * 1e-3
        data["error"] = data.error * 1e-3
        data["unit"] = "keV"
    else:
        raise ValueError(f"Unknown data quantity: {quant}")

    rmag = rmag[0]
    if split == "HFS":
        data = data.where(data.R < rmag)
    elif split == "LFS":
        data = data.where(data.R > rmag)
    else:
        data = data

    return data


def gpr_fit_ts(
    data: xr.DataArray,
    kernel: kernels.Kernel,
    xdim: str = "rho",
    virtual_obs=True,
    virtual_symmetry=True,
    x_bounds=None,
    virtual_points=None,
    plot=False,
    save_fig=False,
):
    if x_bounds is None:
        if xdim == "rho":
            x_bounds = (0, 1.3)
        elif xdim == "R":
            x_bounds = (0.1, 0.9)

    if virtual_points is None:
        if xdim == "rho":
            virtual_points = [
                (1.5, lambda y: 0),
            ]
        elif xdim == "R":
            virtual_points = [(0, lambda y: 0), (0.9, lambda y: 0)]

    x_fit = np.linspace(x_bounds[0], x_bounds[1], 1000)
    y_fit = []
    y_fit_err = []
    gpr = []

    for t in data.t:
        if "t" in data.__getattr__(xdim).dims:
            x = data.__getattr__(xdim).sel(t=t).values
        else:
            x = data.__getattr__(xdim).values
        y = data.sel(t=t).values
        y_err = data.error.sel(t=t).values

        if virtual_obs:
            num_vo = len(virtual_points)
            x = np.insert(
                x,
                [i for i in range(num_vo)],
                [virtual_point[0] for virtual_point in virtual_points],
            )
            y = np.insert(
                y,
                [i for i in range(num_vo)],
                [virtual_point[1](y) for virtual_point in virtual_points],
            )
            y_err = np.insert(
                y_err, [i for i in range(num_vo)], [0.01 for i in range(num_vo)]
            )

            if virtual_symmetry:
                x = np.concatenate((x, -x))
                y = np.concatenate((y, y))
                y_err = np.concatenate((y_err, y_err))

        _y_fit, _y_fit_err, _gpr = gpr_fit(
            x,
            y,
            y_err,
            x_fit,
            kernel,
        )
        y_fit.append(DataArray(_y_fit, coords=[(xdim, x_fit)]))
        y_fit_err.append(DataArray(_y_fit_err, coords=[(xdim, x_fit)]))
        gpr.append(_gpr)

    fit = xr.concat(y_fit, "t").assign_coords(t=data.t)
    fit_err = xr.concat(y_fit_err, "t").assign_coords(t=data.t)

    if plot or save_fig:
        Path(FIG_PATH).mkdir(parents=True, exist_ok=True)
        for idx, tplot in enumerate(data.t.values):
            plot_gpr_fit(
                data.sel(t=tplot).swap_dims({"channel": xdim}),
                fit.sel(t=tplot),
                fit_err.sel(t=tplot),
                ylabel=f"{data.quantity.values} ({data.unit.values})",
                xlabel=f"{xdim}",
                title=f"{data.pulse.values}\nOptimimum: {gpr[idx].kernel_}",
                fig_name=f"{data.pulse.values}_TS.{data.quantity.values}",
                save_fig=save_fig,
            )

            if not save_fig:
                plt.show()

    plt.close("all")
    return fit, fit_err


if __name__ == "__main__":
    pulse = 11089
    tstart = 0.03
    tend = 0.15
    dt = 0.01
    st40 = ReadST40(pulse, tstart, tend, dt)
    st40.get_equilibrium()
    st40(
        instruments=[
            "ts",
            "efit",
        ],
        R_shift=xr.full_like(st40.equilibrium.t, 0.02),
    )

    ne_data = post_process_ts(
        st40.binned_data,
        st40.equilibrium,
        "ne",
        pulse,
        split="",
    )
    ne_kernel = 1.0 * kernels.RationalQuadratic(
        alpha_bounds=(0.1, 1.0), length_scale_bounds=(0.3, 0.7)
    ) + kernels.WhiteKernel(noise_level_bounds=(0.01, 10))

    te_data = post_process_ts(
        st40.binned_data,
        st40.equilibrium,
        "te",
        pulse,
        split="",
    )
    te_kernel = 1.0 * kernels.RationalQuadratic(
        alpha_bounds=(0.1, 1.0), length_scale_bounds=(0.4, 0.7)
    ) + kernels.WhiteKernel(noise_level_bounds=(0.01, 10))

    gpr_fit_ts(
        data=ne_data,
        xdim="rho",
        virtual_obs=True,
        virtual_symmetry=True,
        kernel=ne_kernel,
        save_fig=True,
    )
    gpr_fit_ts(
        data=te_data,
        xdim="rho",
        virtual_obs=True,
        virtual_symmetry=True,
        kernel=te_kernel,
        save_fig=True,
    )



from getpass import getpass
from getpass import getuser
from pathlib import Path
import pickle
import stat
from typing import List
from typing import Optional
from typing import Tuple
import warnings

import numpy as np
from sal.client import SALClient
from sal.core.exception import AuthenticationFailed
from sal.dataclass import Signal

from indica.abstractio import BaseIO
from indica.numpy_typing import RevisionLike
from indica.utilities import CACHE_DIR
from indica.utilities import to_filename


class SALError(Exception):
    """An exception which occurs when trying to read PPF data which would
    not be caught by the lower-level SAL library. An example would be
    failing to find any valid channels for an instrument when each channel
    is a separate DTYPE.

    """


class SALWarning(UserWarning):
    """A warning that occurs while trying to read PPF data. Typically
    related to caching in some way.

    """


class SALUtils(BaseIO):
    def __init__(self, pulse: int, server: str = "https://sal.jet.uk"):
        self.pulse = pulse
        self._reader_cache_id = f"ppf:{server.replace('-', '_')}:{pulse}"
        self._client = SALClient(server)
        self._client.prompt_for_password = False

    def requires_authentication(self) -> None:
        return self._client.auth_required

    def authenticate(self, name: Optional[str] = None) -> bool:
        """Log onto the JET/SAL system to access data.

        Parameters
        ----------
        name:
            Your username when logging onto Heimdall.

        Returns
        -------
        :
            Indicates whether authentication was succesful.
        """
        if name is None:
            name = getuser()
        try:
            self._client.authenticate(name, getpass())
            return True
        except AuthenticationFailed:
            return False

    def close(self) -> None:
        """Ends connection to the SAL server from which PPF data is being
        read."""
        del self._client

    def _get_signal(
        self, uid: str, instrument: str, quantity: str, revision: RevisionLike
    ) -> Tuple[Signal, str]:
        """Gets the signal for the given INSTRUMENT (DDA in JET), at the
        given revision."""
        path = self.get_sal_path(
            uid=uid,
            instrument=instrument,
            revision=self.get_revision(uid, instrument, revision),
            quantity=quantity,
        )
        cache_path = self._sal_path_to_file(path)
        data = self._read_cached_ppf(cache_path)
        if data is None:
            data = self._client.get(path)
            self._write_cached_ppf(cache_path, data)
        return data, path

    def get_signal(
        self, uid: str, instrument: str, quantity: str, revision: RevisionLike
    ) -> Tuple[np.array, str]:
        signal, path = self._get_signal(
            uid=uid,
            instrument=instrument,
            quantity=quantity,
            revision=revision,
        )
        return signal.data, path

    def _get_signal_dims(self, signal: Signal) -> List[np.array]:
        dims = [dim.data for dim in signal.dimensions]
        return dims

    def get_signal_dims(
        self, uid: str, instrument: str, quantity: str, revision: RevisionLike
    ) -> Tuple[List[np.array], List[str]]:
        signal, path = self._get_signal(
            uid=uid,
            instrument=instrument,
            quantity=quantity,
            revision=revision,
        )
        dims = self._get_signal_dims(signal=signal)
        paths = [path] * len(dims)
        return dims, paths

    def _get_signal_units(self, signal: Signal) -> str:
        return signal.units

    def get_signal_units(
        self, uid: str, instrument: str, quantity: str, revision: RevisionLike
    ) -> str:
        signal, path = self._get_signal(
            uid=uid,
            instrument=instrument,
            quantity=quantity,
            revision=revision,
        )
        return self._get_signal_units(signal=signal)

    def get_data(
        self, uid: str, instrument: str, quantity: str, revision: RevisionLike
    ) -> Tuple[np.array, List[np.array], str, str]:
        signal, path = self._get_signal(
            uid=uid,
            instrument=instrument,
            quantity=quantity,
            revision=revision,
        )
        dims = self._get_signal_dims(signal=signal)
        units = self._get_signal_units(signal=signal)
        return signal.data, dims, units, path

    def get_revision(
        self, uid: str, instrument: str, revision: RevisionLike
    ) -> RevisionLike:
        """
        Get actual revision that's being read from database, converts relative revision
        (e.g. 0, latest) to absolute
        """
        info = self._client.list(
            self.get_sal_path(uid=uid, instrument=instrument, revision=revision)
        )
        return info.revision_current

    def _rewrite_instrument(self, instrument: str) -> str:
        """Return modified instrument, if needed, based on input. Used due to slight
        hack in JET PPF system to separate individual instruments that are stored in a
        single PPF.

        """
        instrument_mapping = {
            "sxrh": "sxr",
            "sxrv": "sxr",
            "kb5h": "bolo",
            "kb5v": "bolo",
            "ks3h": "ks3",
            "ks3v": "ks3",
        }
        return instrument_mapping.get(instrument, instrument)

    def get_sal_path(
        self,
        uid: str,
        instrument: str,
        revision: RevisionLike,
        quantity: Optional[str] = None,
    ) -> str:
        """Return the path in the PPF database to for the given INSTRUMENT
        (DDA in JET)."""
        instrument = self._rewrite_instrument(instrument=instrument)
        base_path = f"/pulse/{self.pulse:d}/ppf/signal/{uid}/{instrument}"
        if quantity is not None:
            return base_path + f"/{quantity}:{revision:d}"
        return base_path + f":{revision:d}"

    def _read_cached_ppf(self, path: Path) -> Optional[Signal]:
        """Check if the PPF data specified by `sal_path` has been cached and,
        if so, load it.

        """
        if not path.exists():
            return None
        permissions = stat.filemode(path.stat().st_mode)
        if permissions[5] == "w" or permissions[8] == "w":
            warnings.warn(
                "Can not open cache file which is writeable by anyone other than "
                "the user. (Security risk.)",
                SALWarning,
            )
            return None
        with path.open("rb") as f:
            try:
                return pickle.load(f)
            except pickle.UnpicklingError:
                warnings.warn(
                    f"Error unpickling cache file {path}. (Possible data corruption.)",
                    SALWarning,
                )
                return None

    def _write_cached_ppf(self, path: Path, data: Signal):
        """Write the given signal, fetched from `sal_path`, to the disk for
        later reuse.

        """
        path.parent.mkdir(parents=True, exist_ok=True)
        with path.open("wb") as f:
            pickle.dump(data, f)
        path.chmod(0o644)

    def _sal_path_to_file(self, sal_path: str) -> Path:
        """Get the file path which would be used to cache data from the given
        `sal_path`.

        """
        return (
            Path.home()
            / CACHE_DIR
            / self.__class__.__name__
            / to_filename(self._reader_cache_id + sal_path + ".pkl")
        )



from copy import deepcopy
from typing import Callable
from typing import Dict

import numpy as np
import xarray as xr
from xarray import DataArray

from indica.configs.readers.st40readerprocessorconf import ST40ReaderProcessorConf
from indica.converters.time import convert_in_time_dt


class ReaderProcessor:
    """
    Takes raw data from a datareader and applies filtering and binning
    """

    def __init__(
        self,
        conf=ST40ReaderProcessorConf(),
    ):

        self.conf = conf
        self.raw_data: dict = None
        self.processed_data: dict = {}

    def reset_data(self):
        self.processed_data = {}

    def __call__(
        self,
        raw_data: dict,
        tstart: float = None,
        tend: float = None,
        dt: float = None,
        verbose: bool = False,
    ):

        self.reset_data()
        self.raw_data = raw_data

        self.processed_data = apply_filter(
            raw_data,
            filters=self.conf.filter_values,
            filter_func=value_condition,
            filter_func_name="value",
            verbose=verbose,
        )

        self.processed_data = apply_filter(
            self.processed_data,
            filters=self.conf.filter_coordinates,
            filter_func=coordinate_condition,
            filter_func_name="co-ordinate",
            verbose=verbose,
        )

        self.processed_data = bin_data_in_time(
            self.processed_data, tstart=tstart, tend=tend, dt=dt
        )
        return self.processed_data


def bin_data_in_time(
    raw_data: Dict[str, Dict[str, xr.DataArray]],
    tstart: float = 0.02,
    tend: float = 0.1,
    dt: float = 0.01,
    debug=False,
):
    binned_data = {}
    for instr in raw_data.keys():
        if debug:
            print(f"instr: {instr}")
        binned_quantities = {}
        for quant in raw_data[instr].keys():
            if debug:
                print(f"quant: {quant}")
            data_quant = deepcopy(raw_data[instr][quant])

            if "t" in data_quant.coords:
                data_quant = convert_in_time_dt(tstart, tend, dt, data_quant)
            # Using groupedby_bins always removes error from coords so adding it back
            if "error" in raw_data[instr][quant].coords:
                error = convert_in_time_dt(
                    tstart, tend, dt, raw_data[instr][quant].error
                )
                data_quant = data_quant.assign_coords(
                    error=(raw_data[instr][quant].dims, error.data)
                )
            binned_quantities[quant] = data_quant
        binned_data[instr] = binned_quantities
    return binned_data


def apply_filter(
    data: Dict[str, Dict[str, xr.DataArray]],
    filters: Dict[str, Dict[str, tuple]],
    filter_func: Callable,
    filter_func_name="value",
    verbose=False,
):

    filtered_data = {}
    for instrument, quantities in data.items():
        if instrument not in filters.keys():
            if verbose:
                print(f"missing {filter_func_name} filter for {instrument}")
            filtered_data[instrument] = deepcopy(data[instrument])
            continue

        filtered_data[instrument] = {}
        for quantity_name, quantity in quantities.items():
            if quantity_name not in filters[instrument]:
                filtered_data[instrument][quantity_name] = deepcopy(
                    data[instrument][quantity_name]
                )
                continue

            filter_info = filters[instrument][quantity_name]
            filtered_data[instrument][quantity_name] = filter_func(
                quantity, filter_info
            )
    return filtered_data


def value_condition(data: DataArray, limits: tuple):
    condition = (data >= limits[0]) * (data < limits[1])
    filtered_data = xr.where(condition, data, np.nan)
    filtered_data.attrs = data.attrs
    return filtered_data


def coordinate_condition(data: DataArray, coord_info: tuple):
    coord_name: str = coord_info[0]
    coord_slice: tuple = coord_info[1]
    condition = (data.coords[coord_name] >= coord_slice[0]) * (
        data.coords[coord_name] < coord_slice[1]
    )
    filtered_data = data.where(condition, np.nan)
    filtered_data.attrs = data.attrs
    return filtered_data



from abc import ABC

import matplotlib.pylab as plt


class ProfilerBase(ABC):
    # protocol for profilers to follow

    def __init__(self, parameters: dict = None):
        if parameters is None:
            parameters = {}
        self.parameters = parameters

    def set_parameters(self, **kwargs):
        """
        Set any of the shaping parameters
        """
        for k, v in kwargs.items():
            setattr(self, k, v)
        self.parameters.update(**kwargs)

    def get_parameters(self):
        """
        get all the shaping parameters
        """
        return {key: getattr(self, key) for key in self.parameters.keys()}

    def plot(self, fig=True, **kwargs):
        self.__call__()
        if fig:
            plt.figure()
        self.ydata.plot(**kwargs)

    def __call__(self, *args, **kwargs):
        self.ydata = None



from copy import deepcopy
from typing import Callable
import warnings

import flatdict
from flatdict import FlatDict
import mpmath as mp
import numpy as np

from indica.plasma import PlasmaProfiler

np.seterr(all="ignore")
warnings.simplefilter("ignore", category=FutureWarning)

# using mpmath to handle small numbers as floats insufficient
# wrapping mp functions to process numpy arrays
gaussian = np.frompyfunc(mp.npdf, 3, 1)
log = np.frompyfunc(mp.log, 1, 1)
to_float = np.frompyfunc(float, 1, 1)


def deep_update(mapping: dict, *updating_mappings: dict) -> dict:
    updated_mapping = mapping.copy()
    for updating_mapping in updating_mappings:
        for k, v in updating_mapping.items():
            if (
                k in updated_mapping
                and isinstance(updated_mapping[k], dict)
                and isinstance(v, dict)
            ):
                updated_mapping[k] = deep_update(updated_mapping[k], v)
            else:
                updated_mapping[k] = v
    return updated_mapping


def mp_log_gauss_wrapper(x, mean, sigma):
    # dropping nans to speed up calculation
    nan_idx = np.isnan(x) | np.isnan(mean) | np.isnan(sigma)
    x_star = x[~nan_idx] * mp.mpf(1)  # change dtype to mp float
    mean_star = mean[~nan_idx] * mp.mpf(1)
    sigma_star = sigma[~nan_idx] * mp.mpf(1)
    y = gaussian(x_star, mean_star, sigma_star)
    ln_y = to_float(log(y))  # mpf -> float
    return np.hstack(ln_y[:]).astype(float)  # convert from datatype object to float


class BayesBlackBox:
    """
    Bayesian black box that evaluates an ln_likelihood from given diagnostic data and
    modelled data and an ln_posterior from ln_prior and ln_likelihood

    Parameters
    ----------
    data
        processed diagnostic data of format [diagnostic].[quantity]
    quant_to_optimise
        quantities from data which will be optimised with bckc from diagnostic_models
    plasma_profiler
        plasma interface has methods for setting profiles in plasma
    build_bckc
        function to return model data
    ln_prior
        prior function which calculates ln_prior from given parameters

    """

    def __init__(
        self,
        opt_data: dict,
        quant_to_optimise: list,
        ln_prior: Callable,
        build_bckc: Callable,
        plasma_profiler: PlasmaProfiler,
    ):
        self.opt_data = opt_data
        self.quant_to_optimise = quant_to_optimise
        self.ln_prior = ln_prior
        self.build_bckc = build_bckc

        self.plasma_profiler = plasma_profiler

        missing_data = list(set(quant_to_optimise).difference(opt_data.keys()))
        if missing_data:
            raise ValueError(f"{missing_data} not found in data given")

    def ln_likelihood(self):
        ln_likelihood = 0
        time_coord = self.plasma_profiler.plasma.time_to_calculate

        for key in self.quant_to_optimise:
            model_data = self.bckc[key].values
            exp_data = self.opt_data[key].sel(t=time_coord).values
            exp_error = self.opt_data[key].sel(t=time_coord).error.values

            _ln_likelihood = mp_log_gauss_wrapper(model_data, exp_data, exp_error)
            ln_likelihood += np.nansum(_ln_likelihood)

        return ln_likelihood

    def ln_posterior(self, parameters: dict, **kwargs):
        """
        Posterior probability given to optimisers

        Parameters
        ----------
        parameters
            inputs to optimise
        kwargs
            kwargs for models

        Returns
        -------
        ln_posterior
            log of posterior probability
        blob
           outputs from model bckc and plasma attributes
        """

        _ln_prior = self.ln_prior(parameters)
        if _ln_prior == -np.inf:  # Don't call models if outside priors
            return (
                -1e4,
                {},
            )  # optimisers can't handle inf (1E-10000 is min for numerical precision)

        self.plasma_profiler(parameters)
        plasma_attributes = self.plasma_profiler.plasma_attributes()

        flat_parameters = flatdict.FlatDict(parameters, ".")
        nested_parameters = flat_parameters.as_dict()
        bckc = self.build_bckc(
            **deep_update(kwargs, nested_parameters)
        )  # model calls with optimised params and kwargs
        self.bckc = FlatDict(bckc, ".")

        _ln_likelihood = self.ln_likelihood()  # compare results to data
        ln_posterior = _ln_likelihood + _ln_prior

        blob = deepcopy({**self.bckc, **plasma_attributes})
        return ln_posterior, blob



import matplotlib.cm as cm
import matplotlib.pylab as plt
import numpy as np
import xarray as xr
from xarray import DataArray

from indica.available_quantities import READER_QUANTITIES
from indica.converters import LineOfSightTransform
from indica.models.abstract_diagnostic import AbstractDiagnostic
from indica.numpy_typing import LabeledArray
from indica.operators import PowerLoss
from indica.utilities import build_dataarrays
from indica.utilities import set_axis_sci


class PinholeCamera(AbstractDiagnostic):
    """
    Object representing a pinhole camera diagnostic e.g. bolometer or SXR camera
    """

    def __init__(
        self,
        name: str,
        instrument_method: str = "get_radiation",
        power_loss: dict[str, PowerLoss] = None,
    ):
        self.transform: LineOfSightTransform
        self.name = name
        self.instrument_method = instrument_method
        self.quantities = READER_QUANTITIES[self.instrument_method]
        self.power_loss = power_loss

    def _build_bckc_dictionary(self):
        bckc = {
            "t": self.t,
            "channel": np.arange(len(self.transform.x1)),
            "location": self.transform.origin,
            "direction": self.transform.direction,
            "brightness": self.los_integral,
        }
        self.bckc = build_dataarrays(bckc, self.quantities, transform=self.transform)

    def __call__(
        self,
        Ne: DataArray = None,
        Nion: DataArray = None,
        Lz: dict = None,
        t: LabeledArray = None,
        calc_rho=False,
        sum_beamlets: bool = True,
        **kwargs,
    ):
        """
        Calculate diagnostic measured values

        Parameters
        ----------
        Ne
            Electron density profile (dims = "rho", "t")
        Nion
            Ion density profiles (dims = "rho", "t", "element")
        Lz
            Cooling factor dictionary of DataArrays of each element to be included
        t
            Time (s) for remapping on equilibrium reconstruction

        Returns
        -------
        Dictionary of back-calculated quantities (as datareader.py)

        """

        if self.plasma is not None:
            if t is None:
                t = self.plasma.time_to_calculate
            Ne = self.plasma.electron_density.interp(t=t)
            _Lz = self.plasma.lz_tot
            Lz = {}
            for elem in _Lz.keys():
                Lz[elem] = _Lz[elem].interp(t=t)
            Nion = self.plasma.ion_density.interp(t=t)
        else:
            if Ne is None or Nion is None or Lz is None:
                raise ValueError("Give inputs of assign plasma class!")

        if self.power_loss:
            Lz = {}
            for elem in self.plasma.ion_density.element.values:
                Lz[elem] = self.power_loss[elem](
                    self.plasma.electron_temperature.sel(t=t),
                    self.plasma.fz[elem].sel(t=t).transpose(),
                    Ne=self.plasma.electron_density.sel(t=t),
                    Nh=self.plasma.neutral_density.sel(t=t),
                ).transpose()

        self.t: DataArray = t
        self.Ne: DataArray = Ne
        self.Nion: DataArray = Nion
        self.Lz: dict = Lz

        elements = self.Nion.element.values

        _emissivity = []
        for ielem, elem in enumerate(elements):
            _emissivity.append(
                self.Lz[elem].sum("ion_charge") * self.Nion.sel(element=elem) * self.Ne
            )
        self.emissivity_element = xr.concat(_emissivity, "element")
        self.emissivity = self.emissivity_element.sum("element")

        self.los_integral = self.transform.integrate_on_los(
            self.emissivity,
            t=t,
            calc_rho=calc_rho,
            sum_beamlets=sum_beamlets,
        )

        self._build_bckc_dictionary()

        return self.bckc

    def plot(self, nplot: int = 1):
        if len(self.bckc) == 0:
            print("No model results to plot")
            return

        # Line-of-sight information
        self.transform.plot(np.mean(self.t))

        # Back-calculated profiles
        cols_time = cm.gnuplot2(np.linspace(0.1, 0.75, len(self.t), dtype=float))
        plt.figure()
        for i, t in enumerate(np.array(self.t)):
            if i % nplot:
                continue

            _brightness = self.bckc["brightness"].sel(t=t, method="nearest")
            if "beamlet" in _brightness.dims:
                plt.fill_between(
                    _brightness.channel,
                    _brightness.max("beamlet"),
                    _brightness.min("beamlet"),
                    color=cols_time[i],
                    alpha=0.5,
                )
                brightness = _brightness.mean("beamlet")
            else:
                brightness = _brightness
            brightness.plot(label=f"t={t:1.2f} s", color=cols_time[i])
        set_axis_sci()
        plt.title(self.name.upper())
        plt.xlabel("Channel")
        plt.ylabel("Measured brightness (W/m^2)")
        plt.legend()

        # Local emissivity profiles
        plt.figure()
        for i, t in enumerate(np.array(self.t)):
            if i % nplot:
                continue
            plt.plot(
                self.emissivity.rhop,
                self.emissivity.sel(t=t),
                color=cols_time[i],
                label=f"t={t:1.2f} s",
            )
        set_axis_sci()
        plt.xlabel("rho")
        plt.ylabel("Local radiated power (W/m^3)")
        plt.legend()



"""Routines for getting start and end points of lines of sight using data from
Surf data.

"""

from collections import defaultdict
from pathlib import Path
import re
from typing import DefaultDict
from typing import Iterable
from typing import Optional
from typing import TextIO
from typing import Tuple
from typing import Union

import numpy as np

INSTRUMENT_MAP: DefaultDict[str, Tuple[Optional[str], re.Pattern]] = defaultdict(
    lambda: (None, re.compile(".*")),
    {
        "sxr/h": ("Soft X-ray/KJ5", re.compile(".*")),
        "sxr/t": ("Soft X-ray/KJ3-4 T", re.compile(".*")),
        "sxr/v": ("Soft X-ray/KJ3-4 V", re.compile(".*")),
        "kk3": ("ECE/KK3", re.compile(".*")),
        "bolo/kb5h": ("Bolometry/KB5", re.compile(r"KB5H \d+")),
        "bolo/kb5v": ("Bolometry/KB5", re.compile(r"KB5V \d+")),
    },
)


_DIVIDER = re.compile(r"(?:\s+|\s*,\s*)(?=(?:[^']*'[^']*')*[^']*$)")


class SURFException(Exception):
    """Exception raised when trying to get line of sight data for an
    instrument that does not exist or a pulse number for which data is
    not available.

    """


def _parse_lines(
    data: Iterable[str], criterion: re.Pattern
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """Parse the provided lines of data to get the start and end
    coordinates of the lines of sight. This works when the data is in the
    "lines" format.

    Parameters
    ----------
    data
        An iterable returning lines from the SURF file which should be parsed
        to get line-of-sight data.
    criterion
        A regular expression against which to evaluate the name of each
        channel. The channel will only be included in result if the regular
        expression matches the channel name.

    Retruns
    -------
    Rstart
        Major radius for the start of the line of sight for each channel.
    Rend
        Major radius for the end of the line of sight for each channel.
    Zstart
        Vertical position for the start of the line of sight for each channel.
    Zend
        Vertical position for the end of the line of sight for each channel.
    Tstart
        Toroidal offset of start of the line of sight for each channel.
    Tend
        Toroidal offset of the end of the line of sight for each channel.

    """
    rstart = []
    rend = []
    zstart = []
    zend = []
    for line in data:
        label, rs, zs, re, ze = _DIVIDER.split(line[:-1])
        if criterion.search(label[1:-1]):
            rstart.append(float(rs))
            rend.append(float(re))
            zstart.append(float(zs))
            zend.append(float(ze))
    return (
        np.array(rstart),
        np.array(rend),
        np.array(zstart),
        np.array(zend),
        np.zeros(len(rstart)),
        np.zeros(len(rstart)),
    )


def _parse_line3d(
    data: Iterable[str], criterion: re.Pattern
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """Parse the provided lines of data to get the start and end
    coordinates of the lines of sight. This works when the data is in the
    "line3d" format.

    Parameters
    ----------
    data
        An iterable returning lines from the SURF file which should be parsed
        to get line-of-sight data.
    criterion
        A regular expression against which to evaluate the name of each
        channel. The channel will only be included in result if the regular
        expression matches the channel name.

    Retruns
    -------
    Rstart
        Major radius for the start of the line of sight for each channel.
    Rend
        Major radius for the end of the line of sight for each channel.
    Zstart
        Vertical position for the start of the line of sight for each channel.
    Zend
        Vertical position for the end of the line of sight for each channel.
    Tstart
        Toroidal offset of start of the line of sight for each channel.
    Tend
        Toroidal offset of the end of the line of sight for each channel.

    """
    rstart = []
    rend = []
    zstart = []
    zend = []
    Tstart = []
    Tend = []
    for line in data:
        label, rs, Ts, zs, re, Te, ze = _DIVIDER.split(line[:-1])
        if criterion.search(label[1:-1]):
            rstart.append(float(rs))
            rend.append(float(re))
            zstart.append(float(zs))
            zend.append(float(ze))
            Tstart.append(float(Ts))
            Tend.append(float(Te))
    return (
        np.array(rstart),
        np.array(rend),
        np.array(zstart),
        np.array(zend),
        np.array(Tstart),
        np.array(Tend),
    )


def _parse_kj34(
    data: Iterable[str], criterion: re.Pattern
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """Parse the provided lines of data to get the start and end
    coordinates of the lines of sight. This works when the data is in the
    "kj34" format.

    Parameters
    ----------
    data
        An iterable returning lines from the SURF file which should be parsed
        to get line-of-sight data.
    criterion
        A regular expression against which to evaluate the name of each
        channel. The channel will only be included in result if the regular
        expression matches the channel name.

    Retruns
    -------
    Rstart
        Major radius for the start of the line of sight for each channel.
    Rend
        Major radius for the end of the line of sight for each channel.
    Zstart
        Vertical position for the start of the line of sight for each channel.
    Zend
        Vertical position for the end of the line of sight for each channel.
    Tstart
        Toroidal offset of start of the line of sight for each channel.
    Tend
        Toroidal offset of the end of the line of sight for each channel.

    """
    PIXEL_WIDTH = 0.00099
    indices = np.arange(-17, 18)
    rstart = []
    rend = []
    zstart = []
    zend = []
    for line in data:
        (
            label,
            theta_chip,
            central_index,
            focal_length,
            R_pinhole,
            z_pinhole,
            gamma,
            step,
        ) = _DIVIDER.split(line[:-1])
        if criterion.search(label[1:-1]):
            indices = np.arange(-17, 18, int(step))
            rs = np.empty(len(indices))
            rs[:] = float(R_pinhole) / 1e3
            zs = np.empty(len(indices))
            zs[:] = float(z_pinhole) / 1e3
            f = float(focal_length) / 1e3
            theta = np.radians(float(theta_chip)) + int(gamma) * np.arctan2(
                indices * PIXEL_WIDTH, f
            )
            rstart.append(rs)
            zstart.append(zs)
            rend.append(rs + np.cos(theta))
            zend.append(zs + np.sin(theta))
    rstart = np.concatenate(rstart)  # type: ignore
    rend = np.concatenate(rend)
    zstart = np.concatenate(zstart)  # type: ignore
    zend = np.concatenate(zend)
    Tstart = np.zeros_like(rstart, dtype=float)
    Tend = np.zeros_like(rend, dtype=float)
    return rstart, rend, zstart, zend, Tstart, Tend  # type: ignore


SURF_PARSERS = {
    "lines": _parse_lines,
    "line3d": _parse_line3d,
    "kj34": _parse_kj34,
}


def read_surf_los(
    filename: Union[str, Path], pulse: int, instrument: str
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """Read beginning and ends of lines of sight from Surf data.

    Parameters
    ----------
    filename
        Name of the file containing the FLUSH data.
    pulse
        The pulse number the data is required for.
    instrument
        Which instrument to get the lines of sight for. Format within SURF file
        is inconsistent, but this routine will endeavour to support at least
        the following: SXR/H, SXR/T, SXR/V, KK3, BOLO/KB5H, BOLO/KB5V. For any
        other instrument, try a string which is present within its heading in
        the SURF file.

    Retruns
    -------
    Rstart
        Major radius for the start of the line of sight for each channel.
    Rend
        Major radius for the end of the line of sight for each channel.
    Zstart
        Vertical position for the start of the line of sight for each channel.
    Zend
        Vertical position for the end of the line of sight for each channel.
    Tstart
        Toroidal offset of start of the line of sight for each channel.
    Tend
        Toroidal offset of the end of the line of sight for each channel.
    """
    with open(filename, "r", encoding="latin-1") as f:
        instrument_id, criterion = INSTRUMENT_MAP[instrument.lower()]
        if not instrument_id:
            instrument_id = instrument
        data_format, data = _get_text_block(f, pulse, instrument_id)
    return SURF_PARSERS[data_format](data, criterion)


def _get_text_block(
    file: TextIO, pulse: int, instrument_id: str
) -> Tuple[str, Iterable[str]]:
    """Extract the lines from the SURF file containing the data on lines of
    sight for the requested instrument and pulse number.

    Parameters
    ----------
    file
        A file object from which to extract the data.
    pulse
        The pulse number for which this data is desired.
    instrument_id
        The ID used for the instrument in the SURF file, obtained from
        ``INSTRUMENT_MAP``.

    Returns
    -------
    data_format
        A string indicating how the data is stored and, therefore, which
        function will need to be used to parse it.
    text_lines
        An iterable of the lines of data for the requested instrument/pulse.

    """
    for line in file:
        if line.startswith("*"):
            split = line.split("/")
            if (
                instrument_id in line[1:]
                and pulse >= int(split[-2])
                and pulse <= int(split[-1])
            ):
                break
    else:
        raise SURFException(
            f"File {file.name} has no LOS data for pulse {pulse} and instrument "
            f"{instrument_id}."
        )
    num_lines, columns, data_format = file.readline().split()
    lines = []
    for i in range(int(num_lines)):
        lines.append(file.readline())
    return data_format, lines



from copy import deepcopy
from functools import lru_cache
import hashlib
import pickle
from typing import Callable
from typing import Optional
from typing import Tuple

import numpy as np
import xarray as xr

from indica import Equilibrium
from indica.configs import MACHINE_CONFS
from indica.converters.time import get_tlabels_dt
from indica.numpy_typing import LabeledArray
from indica.operators.atomic_data import default_atomic_data
import indica.physics as ph
from indica.profilers.profiler_base import ProfilerBase
from indica.utilities import format_coord
from indica.utilities import format_dataarray
from indica.utilities import get_element_info


class Plasma:
    def __init__(
        self,
        tstart: float = 0.01,
        tend: float = 0.14,
        dt: float = 0.01,
        machine: str = "st40",
        impurities: Tuple[str, ...] = ("c", "ar"),
        impurity_concentration: Tuple[float, ...] = (0.02, 0.001),  # should be deleted!
        main_ion: str = "h",
        full_run: bool = False,
        n_rad: int = 41,
        n_R: int = 100,
        n_z: int = 100,
        verbose: bool = False,
    ):
        """
        Class for plasma objects.
        - Completely independent of experimental data.
        - Assign an equilibrium object for remapping
        - Independent parameters can be set, dependent ones are properties
        TODO: concentration should not be inputted in initialization!

        tstart
            Start time (s)
        tend
            End time (s)
        dt
            Delta t of time window
        machine
            Machine string identifier
        impurities
            Impurity elements present
        main_ion
            Main ion
        full_run
            If True: compute ionisation balance at every iteration
            If False: calculate default and interpolate
        """
        self.equilibrium: Equilibrium
        self.machine_conf = MACHINE_CONFS[machine]()
        self.tstart = tstart
        self.tend = tend
        self.dt = dt
        self.full_run = full_run
        self.verbose = verbose
        elements: Tuple[str, ...] = (main_ion,)
        for elem in impurities:
            elements += (elem,)
        self.elements = elements
        self.main_ion = main_ion
        self.impurities = impurities
        self.impurity_concentration = impurity_concentration
        self.rho_type = "rhop"

        # Machine attributes
        R0, R1 = self.machine_conf.MACHINE_DIMS[0]
        z0, z1 = self.machine_conf.MACHINE_DIMS[1]
        self.R = format_coord(np.linspace(R0, R1, n_R), "R")
        self.z = format_coord(np.linspace(z0, z1, n_z), "z")

        index = np.arange(n_R)
        R_midplane = np.linspace(self.R.min(), self.R.max(), n_R)
        z_midplane = np.full_like(R_midplane, 0.0)
        coords_midplane = {"index": index}
        self.R_midplane = format_dataarray(R_midplane, "R_midplane", coords_midplane)
        self.z_midplane = format_dataarray(z_midplane, "z_midplane", coords_midplane)

        # Time and radial grid
        self.rhop = format_coord(np.linspace(0, 1.0, n_rad), self.rho_type)
        self.t = format_coord(get_tlabels_dt(self.tstart, self.tend, self.dt), "t")
        self.time_to_calculate = deepcopy(self.t)

        # Elements (ions and specifics of impurities)
        element_z, element_a, element_name, element_symbol = [], [], [], []
        for elem in self.elements:
            _z, _a, _name, _symbol = get_element_info(elem)
            element_z.append(_z)
            element_a.append(_a)
            element_name.append(_name)
            element_symbol.append(_symbol)

        coords_elem = {"element": list(self.elements)}
        self.element_z = format_dataarray(element_z, "atomic_number", coords_elem)
        self.element_a = format_dataarray(element_a, "atomic_weight", coords_elem)
        self.element_name = format_dataarray(element_name, "element_name", coords_elem)
        self.element_symbol = format_dataarray(
            element_symbol, "element_symbol", coords_elem
        )

        # Assign data to variables
        nt = len(self.t)
        nr = len(self.rhop)
        nel = len(self.elements)
        nimp = len(self.impurities)
        data1d_time = np.zeros(nt)
        data2d = np.zeros((nt, nr))
        data2d_elem = np.zeros((nel, nt))
        data3d = np.zeros((nel, nt, nr))
        data3d_imp = np.zeros((nimp, nt, nr))

        coords1d_time = {"t": self.t}
        coords2d = {"t": self.t, self.rho_type: self.rhop}
        coords2d_elem = {"element": list(self.elements), "t": self.t}
        coords3d = {
            "element": list(self.elements),
            "t": self.t,
            self.rho_type: self.rhop,
        }
        coords3d_imp = {
            "element": list(self.impurities),
            "t": self.t,
            self.rho_type: self.rhop,
        }

        # Independent plasma quantities
        self.electron_temperature = format_dataarray(
            data2d, "electron_temperature", coords2d, make_copy=True
        )
        self.electron_density = format_dataarray(
            data2d, "electron_density", coords2d, make_copy=True
        )
        self.neutral_density = format_dataarray(
            data2d, "neutral_density", coords2d, make_copy=True
        )
        self.tau = format_dataarray(data2d, "residence_time", coords2d, make_copy=True)
        self.ion_temperature = format_dataarray(
            data2d, "ion_temperature", coords2d, make_copy=True
        )
        self.toroidal_rotation = format_dataarray(
            data2d, "toroidal_rotation", coords2d, make_copy=True
        )
        self.impurity_density = format_dataarray(
            data3d_imp, "impurity_density", coords3d_imp, make_copy=True
        )
        self.fast_ion_density = format_dataarray(
            data2d, "fast_ion_density", coords2d, make_copy=True
        )
        self.parallel_fast_ion_pressure = format_dataarray(
            data2d, "parallel_fast_ion_pressure", coords2d, make_copy=True
        )
        self.perpendicular_fast_ion_pressure = format_dataarray(
            data2d, "perpendicular_fast_ion_pressure", coords2d, make_copy=True
        )

        # Private variables for class property variables
        self._fast_ion_pressure = format_dataarray(
            data2d, "fast_ion_pressure", coords2d, make_copy=True
        )
        self._electron_pressure = format_dataarray(
            data2d, "electron_pressure", coords2d, make_copy=True
        )
        self._thermal_pressure = format_dataarray(
            data2d, "thermal_pressure", coords2d, make_copy=True
        )
        self._pressure = format_dataarray(data2d, "pressure", coords2d, make_copy=True)
        self._wth = format_dataarray(
            data1d_time, "thermal_stored_energy", coords1d_time, make_copy=True
        )
        self._wfast = format_dataarray(
            data1d_time, "fast_ion_stored_energy", coords1d_time, make_copy=True
        )
        self._wp = format_dataarray(
            data1d_time, "stored_energy", coords1d_time, make_copy=True
        )
        self._zeff = format_dataarray(
            data3d, "effective_charge", coords3d, make_copy=True
        )
        self._ion_density = format_dataarray(
            data3d, "ion_density", coords3d, make_copy=True
        )
        self._meanz = format_dataarray(data3d, "mean_charge", coords3d, make_copy=True)
        self._total_radiation = format_dataarray(
            data3d, "total_radiation", coords3d, make_copy=True
        )
        self._prad_tot = format_dataarray(
            data2d_elem, "total_radiated_power", coords2d_elem, make_copy=True
        )

        _fz = {}
        _lz_tot = {}
        for elem in self.elements:
            nz = self.element_z.sel(element=elem).values + 1
            ion_charge = format_coord(np.arange(nz), "ion_charge")
            coords3d_fract = {
                "t": self.t,
                "rhop": self.rhop,
                "ion_charge": ion_charge,
            }
            data3d_fz = np.full((len(self.t), len(self.rhop), nz), 0.0)
            _fz[elem] = format_dataarray(
                data3d_fz, "fractional_abundance", coords3d_fract, make_copy=True
            )
            _lz_tot[elem] = format_dataarray(
                data3d_fz,
                "total_radiation_loss_parameter",
                coords3d_fract,
                make_copy=True,
            )
        self._fz = _fz
        self._lz_tot = _lz_tot

        # Parameter dependencies relating dependant to independent quantities
        self.Fz = CachedCalculation(
            self.calc_fz,
            [
                self.electron_density,
                self.electron_temperature,
                self.neutral_density,
                self.tau,
            ],
        )

        self.Meanz = CachedCalculation(
            self.calc_meanz,
            [
                self.electron_density,
                self.electron_temperature,
                self.neutral_density,
                self.tau,
            ],
        )

        self.Ion_density = CachedCalculation(
            self.calc_ion_density,
            [
                self.electron_density,
                self.electron_temperature,
                self.impurity_density,
                self.fast_ion_density,
            ],
        )

        self.Zeff = CachedCalculation(
            self.calc_zeff,
            [
                self.electron_density,
                self.electron_temperature,
                self.impurity_density,
                self.fast_ion_density,
                self.neutral_density,
                self.tau,
            ],
        )

        self.Lz_tot = CachedCalculation(
            self.calc_lz_tot,
            [
                self.electron_density,
                self.electron_temperature,
                self.tau,
                self.neutral_density,
            ],
        )

        self.Total_radiation = CachedCalculation(
            self.calc_total_radiation,
            [
                self.electron_density,
                self.electron_temperature,
                self.impurity_density,
                self.fast_ion_density,
                self.tau,
                self.neutral_density,
            ],
        )

        self.Electron_pressure = CachedCalculation(
            self.calc_electron_pressure,
            [
                self.electron_density,
                self.electron_temperature,
            ],
        )

        self.Thermal_pressure = CachedCalculation(
            self.calc_thermal_pressure,
            [
                self.ion_temperature,
                self.electron_density,
                self.electron_temperature,
                self.impurity_density,
                self.fast_ion_density,
            ],
        )

        self.Wth = CachedCalculation(
            self.calc_wth,
            [
                self.ion_temperature,
                self.electron_density,
                self.electron_temperature,
                self.impurity_density,
                self.fast_ion_density,
                # Should include volume but requires equilibrium
            ],
        )

    @property
    def time_to_calculate(self):
        return self._time_to_calculate

    @time_to_calculate.setter
    def time_to_calculate(self, value):
        if np.size(value) == 1:
            self._time_to_calculate = float(value)
        else:
            self._time_to_calculate = np.array(value)

    @property
    def electron_pressure(self):
        return self.Electron_pressure()

    def calc_electron_pressure(self):
        self._electron_pressure.values = ph.calc_pressure(
            self.electron_density, self.electron_temperature
        )
        return self._electron_pressure

    def calc_thermal_pressure(self):
        self._thermal_pressure.values = (
            ph.calc_pressure(self.ion_density, self.ion_temperature).sum("element")
            + self.electron_pressure
        )
        return self._thermal_pressure

    @property
    def thermal_pressure(self):
        return self.Thermal_pressure()

    @property
    def pressure(self):
        self._pressure.values = self.thermal_pressure + self.fast_ion_pressure
        return self._pressure

    @property
    def fast_ion_pressure(self):
        # TODO: check whether degrees of freedom are correctly included...
        self._fast_ion_pressure.values = (
            self.parallel_fast_ion_pressure / 3
            + self.perpendicular_fast_ion_pressure * 2 / 3
        )
        return self._fast_ion_pressure

    def calc_wth(self):
        for t in np.array(self.time_to_calculate, ndmin=1):
            self._wth.loc[dict(t=t)] = (
                3 / 2 * np.trapz(self.thermal_pressure.sel(t=t), self.volume.sel(t=t))
            )
        return self._wth

    @property
    def wth(self):
        return self.Wth()

    @property
    def wfast(self):
        for t in np.array(self.time_to_calculate, ndmin=1):
            self._wfast.loc[dict(t=t)] = (
                3 / 2 * np.trapz(self.fast_ion_pressure.sel(t=t), self.volume.sel(t=t))
            )
        return self._wfast

    @property
    def wp(self):
        return self.wth + self.wfast

    @property
    def fz(self):
        return self.Fz()

    def calc_fz(self):
        for elem in self.elements:
            for t in np.array(self.time_to_calculate, ndmin=1):
                electron_temperature = self.electron_temperature.sel(t=t)
                electron_density = self.electron_density.sel(t=t)
                tau = None
                if np.any(self.tau != 0):
                    tau = self.tau.sel(t=t)
                neutral_density = None
                if np.any(self.neutral_density != 0):
                    neutral_density = self.neutral_density.sel(t=t)
                if any(
                    np.logical_not((electron_temperature > 0) * (electron_density > 0))
                ):
                    continue
                fz_tmp = self.fract_abu[elem](
                    electron_temperature,
                    Ne=electron_density,
                    Nh=neutral_density,
                    tau=tau,
                    full_run=self.full_run,
                )
                self._fz[elem].loc[dict(t=t)] = fz_tmp.transpose()
        return self._fz

    @property
    def zeff(self):
        return self.Zeff()

    def calc_zeff(self):
        self._zeff.values = self.ion_density * self.meanz**2 / self.electron_density
        return self._zeff

    @property
    def ion_density(self):
        return self.Ion_density()
        # return self.calc_ion_density()

    def calc_ion_density(self):
        for elem in self.impurities:
            self._ion_density.loc[dict(element=elem)] = self.impurity_density.sel(
                element=elem
            )

        self._ion_density.loc[dict(element=self.main_ion)] = (
            self.electron_density
            - self.fast_ion_density * self.meanz.sel(element=self.main_ion)
            - (self.impurity_density * self.meanz).sum("element")
        )
        return self._ion_density

    @property
    def lz_tot(self):
        return self.Lz_tot()

    def calc_lz_tot(self):
        fz = self.fz
        for elem in self.elements:
            for t in np.array(self.time_to_calculate, ndmin=1):
                electron_density = self.electron_density.sel(t=t)
                electron_temperature = self.electron_temperature.sel(t=t)
                if any(
                    np.logical_not((electron_temperature > 0) * (electron_density > 0))
                ):
                    continue
                Fz = fz[elem].sel(t=t).transpose()
                neutral_density = None
                if np.any(self.neutral_density.sel(t=t) != 0):
                    neutral_density = self.neutral_density.sel(t=t)
                self._lz_tot[elem].loc[dict(t=t)] = self.power_loss_tot[elem](
                    electron_temperature,
                    Fz,
                    Ne=electron_density,
                    Nh=neutral_density,
                    full_run=self.full_run,
                ).transpose()
        return self._lz_tot

    @property
    def total_radiation(self):
        return self.Total_radiation()

    def calc_total_radiation(self):
        lz_tot = self.lz_tot
        ion_density = self.ion_density
        for elem in self.elements:
            total_radiation = (
                lz_tot[elem].sum("ion_charge")
                * self.electron_density
                * ion_density.sel(element=elem)
            )
            self._total_radiation.loc[dict(element=elem)] = xr.where(
                total_radiation >= 0,
                total_radiation,
                0.0,
            )
        return self._total_radiation

    @property
    def meanz(self):
        return self.Meanz()

    def calc_meanz(self):
        fz = self.fz
        for elem in self.elements:
            self._meanz.loc[dict(element=elem)] = (fz[elem] * fz[elem].ion_charge).sum(
                "ion_charge"
            )
        return self._meanz

    @property
    def prad_tot(self):
        total_radiation = self.total_radiation
        for elem in self.elements:
            for t in np.array(self.time_to_calculate, ndmin=1):
                self._prad_tot.loc[dict(element=elem, t=t)] = np.trapz(
                    total_radiation.sel(element=elem, t=t), self.volume.sel(t=t)
                )
        return self._prad_tot

    @property
    def volume(self):
        return self.equilibrium.volume.interp(rhop=self.rhop, t=self.t)

    @property
    def area(self):
        return self.equilibrium.area.interp(rhop=self.rhop, t=self.t)

    @property
    def rmjo(self):
        return self.equilibrium.rmjo.interp(rhop=self.rhop, t=self.t)

    @property
    def rmji(self):
        return self.equilibrium.rmji.interp(rhop=self.rhop, t=self.t)

    @property
    def rmag(self):
        return self.equilibrium.rmag.interp(t=self.t)

    @property
    def zmag(self):
        return self.equilibrium.zmag.interp(t=self.t)

    @property
    def rmin(self):
        return (self.rmjo - self.rmji) / 2.0

    def set_impurity_concentration(
        self,
        element: str,
        concentration: float,
        t: Optional[LabeledArray] = None,
        flat_zeff: bool = False,
    ):
        """
        Sets impurity density for a specific element = concentration * electron_density

        Parameters
        ----------
        element
            string impurity identifier
        concentration
            value of the desired concentration
        t
            time for which concentration is to be set
        flat_zeff
            if True, modifies impurity density to get a ~ flat zeff contribution
        """
        if t is None:
            t = self.time_to_calculate

        if element in self.elements:
            el_dens = self.electron_density.sel(t=t)
            _imp_dens = el_dens * concentration
            if flat_zeff and np.count_nonzero(_imp_dens) != 0:
                meanz = self.meanz.sel(element=element).sel(t=t)
                _zeff = _imp_dens * meanz**2 / el_dens
                zeff_core = _zeff.where(self.rhop < 0.5).mean("rhop")
                imp_dens = el_dens * zeff_core / meanz**2
            else:
                imp_dens = _imp_dens

            self.impurity_density.loc[dict(element=element, t=t)] = imp_dens.values

    def set_equilibrium(self, equilibrium: Equilibrium):
        """Assign equilibrium object and associated private variables"""
        self.equilibrium = equilibrium

    def set_adf11(self, adf11: dict):
        self.adf11 = adf11

    def build_atomic_data(self):
        """
        Assigns default atomic fractional abundance and radiated power operators
        """
        fract_abu, power_loss_tot = default_atomic_data(self.elements)
        self.fract_abu = fract_abu
        self.power_loss_tot = power_loss_tot

    # TODO: if ion asymmetry parameters are not == 0, calculate 2D (R, z) maps
    def map_to_2d(self):
        """
        Calculate total radiated power on a 2D poloidal plane
        including effects from poloidal asymmetries
        """
        print("\n Not implemented yet")

    def write_to_pickle(self, pulse: int = None):
        with open(f"data_{pulse}.pkl", "wb") as f:
            pickle.dump(
                self,
                f,
            )


# Generalized dependency caching
class TrackDependecies:
    def __init__(
        self,
        operator: Callable,
        dependencies: list,
    ):
        """
        Call operator only if dependencies variables have changed.
        Currently using np array, can change if needed

        Parameters
        ----------
        operator
            Function to be called
        dependencies
            Tuple of variables to be tracked
        """
        self.operator = operator
        self.dependencies = dependencies

    def numpyhash(
        self,
        nparray: np.array,
    ):
        a = nparray.view(np.uint8)
        return hashlib.sha1(a).hexdigest()

    def __hash__(self):
        """
        Hashing of dependencies for caching

        xr.DataArray, and np.ndarray are the only types currently permitted.
        Dictionaries contained in dependencies were not updating to match the
        plasma attributes and so were removed.

        TODO: upgrade so other objects can be hashed, e.g. Equilibrium
        """
        _dependencies = []
        for dependency in self.dependencies:
            if type(dependency) == dict:
                raise NotImplementedError(
                    "dictionary dependencies are not working correctly"
                )
            elif type(dependency) == xr.DataArray:
                _dependencies.append(dependency.data)
            elif type(dependency) == np.ndarray:
                _dependencies.append(dependency)
            else:
                raise NotImplementedError(
                    f"Hashing only implemented for xr.DataArray and np.ndarray "
                    f"not {type(dependency)}"
                )

        hashable = tuple((self.numpyhash(dependency),) for dependency in _dependencies)
        return hash(hashable)


class CachedCalculation(TrackDependecies):
    def __init__(self, operator: Callable, dependencies: list, verbose: bool = False):
        self.verbose = verbose
        super(CachedCalculation, self).__init__(operator, dependencies)

    @lru_cache()
    def __call__(self):
        if self.verbose:
            print("Calculating")
        return deepcopy(self.operator())


class PlasmaProfiler:
    def __init__(
        self,
        plasma: Plasma,
        profilers: dict[ProfilerBase],
        plasma_attribute_names=None,
        map_vtor: bool = False,
    ):
        """
        Interface Profiler objects with Plasma object to generate plasma profiles
        and update them.

        Parameters
        ----------
        plasma
            Plasma object
        profilers
            dictionary of Profiler objects to generate profiles
        """

        if plasma_attribute_names is None:
            plasma_attribute_names = [
                "electron_temperature",
                "electron_density",
                "ion_temperature",
                "ion_density",
                "impurity_density",
                "fast_ion_density",
                "fast_ion_pressure",
                "neutral_density",
                "zeff",
                "meanz",
                "wp",
                "wth",
                "pressure",
                "thermal_pressure",
                "toroidal_rotation",
            ]
        self.plasma = plasma
        self.profilers = profilers
        self.plasma_attribute_names = plasma_attribute_names
        self.map_vtor = map_vtor
        self.phantom = None
        self.phantom_profiles = None

    def update_profilers(self, profilers: dict):
        for profile_name, profiler in profilers.items():
            self.profilers[profile_name] = profiler

    def set_profiles(self, profiles: dict[xr.DataArray], t: float = None):
        if t is None:
            t = self.plasma.time_to_calculate

        for profile_name, profile in profiles.items():
            _prof_identifiers = profile_name.split(
                ":"
            )  # impurities have ':' to identify elements
            if profile_name.__contains__(":"):
                if _prof_identifiers[1] in self.plasma.elements:
                    getattr(self.plasma, _prof_identifiers[0]).loc[
                        dict(t=t, element=_prof_identifiers[-1])
                    ] = profile
                else:
                    print(
                        f"profile {profile_name} can't be set because "
                        f"{_prof_identifiers[1]} not in plasma.elements"
                    )
            else:
                getattr(self.plasma, profile_name).loc[dict(t=t)] = profile

    def save_phantoms(self, phantom=False):
        #  if phantoms return profiles otherwise return empty arrays
        self.phantom = phantom
        phantom_profiles = {"PSI_NORM": self.plasma_attributes()}
        if not phantom:
            for key, value in phantom_profiles["PSI_NORM"].items():
                phantom_profiles["PSI_NORM"][key] = value * 0
        phantom_profiles["R_MIDPLANE"] = self.map_plasma_profile_to_midplane(
            phantom_profiles["PSI_NORM"]
        )
        self.phantom_profiles = phantom_profiles
        return phantom_profiles

    def map_plasma_profile_to_midplane(self, profiles: dict):
        """
        Map profiles from flux space to real space on z=0
        """
        midplane_profiles: dict = {}

        R = self.plasma.R_midplane
        z = self.plasma.z_midplane
        _rhop, _, _ = self.plasma.equilibrium.flux_coords(R, z, self.plasma.t)
        rhop = _rhop.swap_dims({"index": "R"})

        for key, value in profiles.items():
            if "rhop" not in value.dims:
                continue
            if not hasattr(self.plasma, key):
                continue
            midplane_profiles[key] = value.interp(rhop=rhop)
        return midplane_profiles

    def plasma_attributes(self):
        plasma_attributes = {}
        for attribute in self.plasma_attribute_names:
            plasma_attributes[attribute] = getattr(self.plasma, attribute).sel(
                t=self.plasma.time_to_calculate
            )
        return plasma_attributes

    def map_toroidal_rotation_to_ion_temperature(
        self,
    ):

        self.plasma.toroidal_rotation = (
            self.plasma.ion_temperature
            / self.plasma.ion_temperature.max("rhop")
            * self.plasma.toroidal_rotation.max("rhop")
        )

    def __call__(self, parameters: dict = None, t=None):
        """
        Set parameters of given profilers and assign to plasma profiles

        parameters
            Flat dictionary of {"profile_name.param_name":value}
            Special case for impurity density:
                {"profile_name:element.param_name":value}
        t
            time points to apply changes
        """
        if parameters is None:
            parameters = {}

        _profiles_to_update: list = []

        # set params for all profilers
        for parameter_name, parameter in parameters.items():
            profile_name, profile_param_name = parameter_name.split(".")
            if profile_name not in self.profilers.keys():
                continue

            if not hasattr(self.profilers[profile_name], profile_param_name):
                raise ValueError(
                    f"No parameter {profile_param_name} available for {profile_name}"
                )
            self.profilers[profile_name].set_parameters(
                **{profile_param_name: parameter}
            )
            _profiles_to_update.append(profile_name)

        # Update only desired profiles or if no parameters given update all
        if _profiles_to_update:
            profiles_to_update = list(set(_profiles_to_update))
        else:
            print("no profile params given so updating all")
            profiles_to_update = list(self.profilers.keys())

        updated_profiles = {
            profile_to_update: self.profilers[profile_to_update]()
            for profile_to_update in profiles_to_update
        }
        self.set_profiles(updated_profiles, t)



"""Definition of weighting functions for line of sight spots.
"""

# import matplotlib.pylab as plt
import numpy as np

from .line_of_sight import LineOfSightTransform

# import xarray as xr


class SpotWeightings:
    """ """

    def __init__(
        self,
        los_transform: LineOfSightTransform,
        dist_func: str,
        sigma_w: float = 0.01,
        sigma_v: float = 0.01,
        p_w: float = 1.0,
        p_v: float = 1.0,
    ):
        self.los_transform = los_transform
        self.dist_func = dist_func
        self.sigma_w = sigma_w
        self.sigma_v = sigma_v
        self.p_w = p_w
        self.p_v = p_v

        n_w = int(np.sqrt(los_transform.beamlets))
        n_v = int(np.sqrt(los_transform.beamlets))
        grid_w = np.linspace(
            -los_transform.spot_width / 2,
            los_transform.spot_width / 2,
            n_w * 2 + 1,
            dtype=float,
        )
        grid_w = grid_w[1::2]
        grid_v = np.linspace(
            -los_transform.spot_height / 2,
            los_transform.spot_height / 2,
            n_v * 2 + 1,
            dtype=float,
        )
        grid_v = grid_v[1::2]
        W, V = np.meshgrid(grid_w, grid_v)

        self.delta_w = grid_w
        self.delta_v = grid_v
        self.W = W
        self.V = V

        if dist_func.lower() == "gaussian":
            print("GAUSSIAN DRAGON")
            self.super_gaussian()

        else:
            raise ValueError("dist_func does not exist")

    def super_gaussian(self):
        self.weightings = np.exp(
            -(((self.W) ** 2 / (2 * self.sigma_w**2)) ** self.p_w)
            - ((self.V) ** 2 / (2 * self.sigma_v**2)) ** self.p_v
        )



from copy import deepcopy

import numpy as np
from scipy.interpolate import CubicSpline
import xarray as xr
from xarray import DataArray

from indica.profilers.profiler_gauss import ProfilerGauss
from indica.utilities import set_plot_colors

CMAP, COLORS = set_plot_colors()


def sawtooth_crash(pre: ProfilerGauss, rho_inv: float, volume: DataArray = None):
    """
    Model a sawtooth crash, without resetting the internal quantities

    If volume in not None, then volume integral is conserved adapting the edge shape

    Parameters
    ----------
    rho_inv
        Inversion radius
    volume
        Plasma volume

    """
    post = deepcopy(pre)

    if volume is None:
        volume = DataArray(0.85 * pre.xspl**3, coords=[("rhop", pre.xspl)])
    vol = volume.interp(rhop=pre.ydata.rhop)
    vol_int_pre = np.trapz(pre.ydata, vol)

    x = pre.x[np.where(pre.x <= 1.0)[0]]

    rho = post.ydata.rhop
    inv_ind = np.max(np.where(rho <= rho_inv)[0])
    for rind in np.arange(inv_ind, rho.size):
        y = xr.where(rho <= rho[rind], post.ydata.sel(rhop=rho[inv_ind]), post.ydata)
        vol_int_post = np.trapz(y, vol)
        if vol_int_post >= vol_int_pre:
            break

    y = xr.where(rho != rho[rind], y, (y[rind] + y[rind + 1]) / 2)
    y = y.interp(rhop=x)

    x = np.append(x, pre.xend)
    y = np.append(y, pre.yend)
    cubicspline = CubicSpline(
        x,
        y,
        0,
        "clamped",
        False,
    )
    post.ydata.values = cubicspline(pre.xspl)
    vol_int_post = np.trapz(post.ydata, vol)

    print(f"Vol-int: {vol_int_pre}, {vol_int_post}")

    return post



"""Experimental design for reading data from disk/database."""

from abc import ABC
from typing import Any
from typing import Dict
from typing import Tuple

import numpy as np
from xarray import DataArray

from indica import BaseIO
from indica.available_quantities import READER_QUANTITIES
from indica.configs.readers.machineconf import MachineConf
from indica.converters import CoordinateTransform
from indica.numpy_typing import RevisionLike
from indica.utilities import build_dataarrays


class DataReader(ABC):
    """Abstract base class to read data in from a database."""

    def __init__(
        self,
        pulse: int,
        tstart: float,
        tend: float,
        machine_conf: MachineConf,
        reader_utils: BaseIO,
        verbose: bool = False,
        **kwargs: Any,
    ):
        """
        Parameters
        ----------
        tstart
            Start of time range for which to get data.
        tend
            End of time range for which to get data.
        kwargs
            Any other arguments which should be recorded for the reader.
        """
        self.verbose = verbose
        self.pulse = pulse
        self.tstart = tstart
        self.tend = tend
        self.reader_utils = reader_utils
        self.machine_conf = machine_conf()
        self.instrument_methods = self.machine_conf.INSTRUMENT_METHODS
        self.machine_dims = self.machine_conf.MACHINE_DIMS
        self.quantities_path = self.machine_conf.QUANTITIES_PATH
        self.verbose = verbose
        self.kwargs = kwargs

    def get(
        self,
        uid: str,
        instrument: str,
        revision: RevisionLike = 0,
        dl: float = 0.005,
        passes: int = 1,
        include_error: bool = True,
        return_dataarrays: bool = True,
        verbose: bool = False,
    ) -> Dict[str, DataArray]:
        """General method that reads data for a requested instrument."""
        if instrument not in self.instrument_methods.keys():
            raise ValueError(
                "{} does not support reading for instrument {}".format(
                    self.__class__.__name__, instrument
                )
            )

        # Read data from database
        _database_results = self._read_database(uid, instrument, revision)
        _database_results["dl"] = dl
        _database_results["passes"] = passes

        # Re-arrange data (machine-specific) and get instrument geometry transform
        method = self.instrument_methods[instrument]
        database_results, transform = getattr(self, f"_{method}")(_database_results)
        if not return_dataarrays:
            return database_results

        quantities = READER_QUANTITIES[method]
        data_arrays = build_dataarrays(
            database_results,
            quantities,
            self.tstart,
            self.tend,
            transform,
            include_error,
            verbose=verbose,
        )
        return data_arrays

    def _read_database(
        self,
        uid: str,
        instrument: str,
        revision: RevisionLike,
    ) -> dict:
        """Read and return all raw database quantities and errors
        Exception handling is non-specific to guarantee generality across readers.

        Include in data dictionary also the UID, INSTRUMENT, MACHINE_DIMS and REVISION
        to guarantee data traceability across data-structures.

        TODO: move error/dimensions/units/records to sub-dictionary within results e.g.
              results = {..., "error":{}, "dimensions":{}, "units":{}, "records":{}}
        """
        method = self.instrument_methods[instrument]
        quantities_paths = self.quantities_path[method]

        revision = self.reader_utils.get_revision(uid, instrument, revision)
        results: Dict[str, Any] = {
            "uid": uid,
            "instrument": instrument,
            "machine_dims": self.machine_dims,
            "revision": revision,
        }
        for _key, _path in quantities_paths.items():
            _key_err = _key + "_error"
            _path_err = _path + "_err"

            # Read quantity value
            try:
                q_val, q_dimensions, q_units, q_path = self.reader_utils.get_data(
                    uid,
                    instrument,
                    _path,
                    revision,
                )
            except Exception as e:
                if self.verbose:
                    print(f"Error reading {_path}: {e}")
                    raise e
                continue
            results[_key + "_records"] = q_path
            results[_key + "_dimensions"] = q_dimensions
            results[_key + "_units"] = q_units
            results[_key] = q_val

            # Read quantity error
            try:
                (
                    q_err,
                    q_err_dimensions,
                    q_err_units,
                    q_err_path,
                ) = self.reader_utils.get_data(
                    uid,
                    instrument,
                    _path_err,
                    revision,
                )
            except Exception as e:
                q_err = np.full_like(results[_key], 0.0)
                q_err_dimensions = []
                q_err_units = ""
                q_err_path = f"{e}"
            results[_key_err] = q_err
            results[_key_err + "_records"] = q_err_path
            results[_key_err + "_dimensions"] = q_err_dimensions
            results[_key_err + "_units"] = q_err_units

        return results

    # Machine-specific instrument methods that must be implemented in the child reader
    # to refactor database data structures and assign a geometry transform
    def _get_thomson_scattering(
        self,
        data: dict,
    ) -> Tuple[Dict[str, Any], CoordinateTransform]:
        raise NotImplementedError

    def _get_profile_fits(
        self,
        data: dict,
    ) -> Tuple[Dict[str, Any], CoordinateTransform]:
        raise NotImplementedError

    def _get_charge_exchange(
        self,
        data: dict,
    ) -> Tuple[Dict[str, Any], CoordinateTransform]:
        raise NotImplementedError

    def _get_spectrometer(
        self,
        data: dict,
    ) -> Tuple[Dict[str, Any], CoordinateTransform]:
        raise NotImplementedError

    def _get_equilibrium(
        self,
        data: dict,
    ) -> Tuple[Dict[str, Any], CoordinateTransform]:
        raise NotImplementedError

    def _get_radiation(
        self,
        data: dict,
    ) -> Tuple[Dict[str, Any], CoordinateTransform]:
        raise NotImplementedError

    def _get_helike_spectroscopy(
        self,
        data: dict,
    ) -> Tuple[Dict[str, Any], CoordinateTransform]:
        raise NotImplementedError

    def _get_diode_filters(
        self,
        data: dict,
    ) -> Tuple[Dict[str, Any], CoordinateTransform]:
        raise NotImplementedError

    def _get_interferometry(
        self,
        data: dict,
    ) -> Tuple[Dict[str, Any], CoordinateTransform]:
        raise NotImplementedError

    def _get_zeff(
        self,
        data: dict,
    ) -> Tuple[Dict[str, Any], CoordinateTransform]:
        raise NotImplementedError



from copy import deepcopy

from matplotlib import cm
import matplotlib.pylab as plt
import numpy as np
import xarray as xr
from xarray import DataArray

from indica.numpy_typing import LabeledArray
from indica.profilers.profiler_gauss import ProfilerGauss
from indica.utilities import FIG_PATH
from indica.utilities import save_figure
from indica.utilities import set_axis_sci
from indica.utilities import set_plot_rcparams

CMAP = cm.gnuplot2

PARAMETER_LIMITS: dict = {
    "electron_temperature": {
        "y0": (500, 10000),
        "y1": (30, 100),
        "peaking": (1, 5),
        "wcenter": (0.2, 0.4),
        "wped": (2, 10),
    },
    "ion_temperature": {
        "y0": (500, 10000),
        "y1": (30, 100),
        "peaking": (1, 5),
        "wcenter": (0.2, 0.4),
        "wped": (2, 10),
    },
    "electron_density": {
        "y0": (1.0e19, 1.0e20),
        "y1": (1.0e18, 5.0e18),
        "peaking": (1, 3),
        "wcenter": (0.3, 0.4),
        "wped": (2, 10),
    },
    "thermal_neutral_density": {
        "y0": (1.0e13, 1.0e14),
        "y1": (1.0e15, 1.0e16),
        "yend": "y1",
        "peaking": (1, 1),
        "wcenter": (0, 0),
        "wped": (5, 15),
    },
}


def profile_scans_pca(
    parameter_limits: dict = None,
    scans: int = 100,
    datatype: str = "electron_temperature",
    rho: LabeledArray = np.linspace(0, 1.0, 41),
    plot: bool = True,
    save_fig: bool = False,
    fig_path: str = None,
):
    if fig_path is None:
        fig_path = FIG_PATH

    set_plot_rcparams("profiles")

    if parameter_limits is None:
        parameter_limits = PARAMETER_LIMITS[datatype]

    profiler = ProfilerGauss(datatype=datatype, xspl=rho)

    _profile_scans: list = []
    parameter_scans: dict = {}
    parameter_keys = parameter_limits.keys()
    for k in parameter_keys:
        parameter_scans[k] = []

    for iscan in range(scans):
        for k, lim in parameter_limits.items():
            if type(lim) != str:
                param_value = np.random.uniform(lim[0], lim[1])
            else:
                param_value = parameter_scans[lim][-1]

            parameter_scans[k].append(param_value)
            setattr(profiler, k, param_value)
        _profile_scans.append(profiler())

    profile_scans: DataArray = xr.concat(_profile_scans, "scan").assign_coords(
        {"scan": np.arange(scans)}
    )

    for k in parameter_keys:
        profile_scans = profile_scans.assign_coords({k: ("scan", parameter_scans[k])})

    if plot:
        cols = CMAP(np.linspace(0.1, 0.75, scans, dtype=float))
        plt.figure()
        sort_ind = np.argsort(profile_scans.sel(rhop=0)).values
        for i, scan in enumerate(sort_ind):
            profile_scans.sel(scan=scan).plot(
                alpha=0.6,
                color=cols[i],
            )
        plt.title("")
        if datatype == "thermal_neutral_density":
            plt.yscale("log")
        set_axis_sci()
        save_figure(fig_path, f"{datatype}_uniform_scan", save_fig=save_fig)

        for k in parameter_keys:
            counts, bins = np.histogram(getattr(profile_scans, k))
            plt.figure()
            plt.stairs(counts, bins)
            plt.title(f"{k} distribution")

    return profile_scans, profiler


def profile_scans_hda(plot=False, rhop=np.linspace(0, 1.0, 41)):
    Te = ProfilerGauss(datatype="electron_temperature", xspl=rhop)
    Ne = ProfilerGauss(datatype="electron_density", xspl=rhop)
    Nimp = ProfilerGauss(datatype="impurity_density", xspl=rhop)
    Vrot = ProfilerGauss(datatype="toroidal_rotation", xspl=rhop)

    Te_list = {}
    Ti_list = {}
    Ne_list = {}
    Nimp_list = {}
    Vrot_list = {}

    # Broad Te profile
    Te.y1 = 30
    Te.wped = 3
    Te.wcenter = 0.35
    Te()
    Te_list["broad"] = deepcopy(Te)
    if plot:
        plt.figure()
        Te.ydata.plot(color="black", label="Te")

    # Broad Ti profile without/with Te as reference
    Ti = deepcopy(Te)
    Ti.datatype = "ion_temperature"
    Ti.y0 = 7.0e3
    Ti()
    Ti_list["broad"] = deepcopy(Ti)
    if plot:
        Ti.ydata.plot(linestyle="dashed", color="black", label="Ti no ref")
    Ti(y0_ref=Te.ydata.sel(rhop=0).values)
    if plot:
        Ti.ydata.plot(linestyle="dotted", color="black", label="Ti with ref")

    # Peaked Te profile
    Te.wcenter, Te.wped, Te.y1, Te.peaking = (0.35, 2, 10, 4)
    Te()
    Te_list["peaked"] = deepcopy(Te)
    if plot:
        Te.ydata.plot(color="red", label="Te")

    # Peaked Ti profile without/with Te as reference
    Ti = deepcopy(Te)
    Ti.datatype = "ion_temperature"
    Ti.y0 = 5.0e3
    Ti()
    Ti_list["peaked"] = deepcopy(Ti)
    if plot:
        Ti.ydata.plot(linestyle="dashed", color="red", label="Ti no ref")
    Ti(y0_ref=Te.ydata.sel(rhop=0).values)
    if plot:
        Ti.ydata.plot(linestyle="dotted", color="red", label="Ti with ref")

    Ne.wped = 6
    Ne.y1 = 0.5e19
    Ne()
    Ne_list["broad"] = deepcopy(Ne)
    if plot:
        plt.figure()
        Ne.ydata.plot(color="black")

    Ne.wped = 3.5
    Ne.peaking = 4
    Ne.y1 = 0.1e19
    Ne()
    Ne_list["peaked"] = deepcopy(Ne)
    if plot:
        Ne.ydata.plot(color="red")

    Nimp.wped = 6
    Nimp.y0 = 5.0e16
    Nimp.y1 = 3.0e16
    Nimp.yend = 2.0e16
    Nimp()
    Nimp_list["flat"] = deepcopy(Nimp)
    if plot:
        plt.figure()
        Nimp.ydata.plot(color="black")

    Nimp.peaking = 8
    Nimp.wcenter = 0.2
    Nimp.y1 = 0.5e16
    Nimp.yend = 0.5e16
    Nimp()
    Nimp_list["peaked"] = deepcopy(Nimp)
    if plot:
        Nimp.ydata.plot(color="red")

    Vrot.y1 = 1.0e3
    Vrot.yend = 0.0
    Vrot()
    Vrot_list["broad"] = deepcopy(Vrot)
    if plot:
        plt.figure()
        Vrot.ydata.plot(color="black")

    Vrot.wped = 1
    Vrot.peaking = 2.0
    Vrot()
    Vrot_list["peaked"] = deepcopy(Vrot)
    if plot:
        Vrot.ydata.plot(color="red")

    return {
        "Te": Te_list,
        "Ti": Ti_list,
        "Ne": Ne_list,
        "Nimp": Nimp_list,
        "Vrot": Vrot_list,
    }

    # import pandas as pd
    # to_write = {
    #     "Rho-poloidal": rho,
    #     "Te broad (eV)": Te_broad.ydata.values,
    #     "Te peaked (eV)": Te_peak.ydata.values,
    #     "Ti broad (eV)": Ti_broad.ydata.values,
    #     "Ti peaked (eV)": Ti_peak.ydata.values,
    #     "Ne broad (m^-3)": Ne_broad.ydata.values,
    #     "Ne peaked (m^-3)": Ne_peak.ydata.values,
    # }
    # df = pd.DataFrame(to_write)
    # df.to_csv("/home/marco.sertoli/data/Indica/profiles.csv")



"""Base class for reading in ADAS atomic data."""

import datetime
from pathlib import Path
import re
from typing import List
from typing import Literal
from typing import TextIO
from typing import Union
from urllib.request import pathname2url
from urllib.request import urlretrieve

import numpy as np
from xarray import DataArray

from indica import BaseIO
from indica.utilities import assign_datatype
from indica.utilities import CACHE_DIR

# TODO: Evaluate this location
DEFAULT_PATH = Path("")

ADF11: dict = {
    "h": {
        "scd": "96",
        "acd": "96",
        "ccd": "96",
        "plt": "96",
        "prb": "96",
        "prc": "96",
        "pls": "15",
        "prs": "15",
    },
    "he": {
        "scd": "89",
        "acd": "89",
        "ccd": "89",
        "plt": "89",
        "prb": "89",
        "prc": "89",
        "pls": "15",
        "prs": "15",
    },
    "li": {
        "scd": "89",
        "acd": "89",
        "ccd": "89",
        "plt": "89",
        "prb": "89",
        "prc": "89",
        "pls": "15",
        "prs": "15",
    },
    "be": {
        "scd": "89",
        "acd": "89",
        "ccd": "89",
        "plt": "89",
        "prb": "89",
        "prc": "89",
        "pls": "15",
        "prs": "15",
    },
    "c": {
        "scd": "89",
        "acd": "89",
        "ccd": "89",
        "plt": "89",
        "prb": "89",
        "prc": "89",
        "pls": "15",
        "prs": "15",
    },
    "ne": {
        "scd": "89",
        "acd": "89",
        "ccd": "89",
        "plt": "89",
        "prb": "89",
        "prc": "89",
        "pls": "15",
        "prs": "15",
    },
    "ar": {
        "scd": "89",
        "acd": "89",
        "ccd": "89",
        "plt": "42",
        "prb": "89",
        "prc": "89",
        "pls": "15",
        "prs": "15",
    },
    "ni": {
        "scd": "89",
        "acd": "89",
        "ccd": "89",
        "plt": "89",
        "prb": "89",
        "prc": "89",
        "pls": "15",
        "prs": "15",
    },
    "kr": {
        "scd": "89",
        "acd": "89",
        "ccd": "89",
        "plt": "42",
        "prb": "89",
        "prc": "89",
        "pls": "15",
        "prs": "15",
    },
    "xe": {
        "scd": "89",
        "acd": "89",
        "ccd": "89",
        "plt": "42",
        "prb": "89",
        "prc": "89",
        "pls": "15",
        "prs": "15",
    },
    "mo": {
        "scd": "89",
        "acd": "89",
        "ccd": "89",
        "plt": "89",
        "prb": "89",
        "prc": "89",
    },
    "w": {
        "scd": "89",
        "acd": "89",
        "ccd": "89",
        "plt": "42",
        "prb": "89",
        "prc": "89",
        "pls": "15",
        "prs": "15",
    },
}


class ADASReader(BaseIO):
    """Class for reading atomic data from ADAS files.

    Parameters
    ----------
    path: Union[str, Path]
        Location from which relative paths should be evaluated.
        Default is to download files from OpenADAS, storing them
        in your home directory for later use.
    """

    def __init__(
        self,
        path: Union[str, Path] = DEFAULT_PATH,
    ):
        path = Path(path)
        self.openadas = path == DEFAULT_PATH
        if path == DEFAULT_PATH:
            self.namespace = "openadas"
            self.path = Path.home() / CACHE_DIR / "adas"
        else:
            self.namespace = "localadas"
            self.path = path

    def close(self):
        """Closes connections to database/files. For this class it does not
        need to do anything."""
        pass

    def get_adf11(self, quantity: str, element: str, year: str) -> DataArray:
        """Read data from the specified ADAS file.

        Parameters
        ----------
        quantity
            The type of data to retrieve. Options: scd, acd, plt, prb,
            plsx, prsx.
        element
            The atomic symbol for the element which will be retrieved.
        year
            The two-digit year label for the data.

        Returns
        -------
        :
            The data in the specified file. Dimensions are density and
            temperature. Each members of the dataset correspond to a
            different charge state.

        """
        now = datetime.datetime.now()
        file_component = f"{quantity}{year}"
        filename = Path(file_component) / f"{file_component}_{element.lower()}.dat"
        with self._get_file("adf11", filename) as f:
            header = f.readline().split()
            # z = int(header[0])
            nd = int(header[1])
            nt = int(header[2])
            zmin = int(header[3]) - 1
            zmax = int(header[4]) - 1
            element_name = header[5][1:].lower()
            f.readline()
            densities = np.fromfile(f, float, nd, " ")
            temperatures = np.fromfile(f, float, nt, " ")
            data = np.empty((zmax - zmin + 1, nt, nd))
            date = datetime.date.min
            for i in range(zmax - zmin + 1):
                section_header = f.readline()
                m = re.search(r"Z1=\s*(\d+)", section_header, re.I)
                assert isinstance(m, re.Match)
                assert int(m.group(1)) - 1 == zmin + i
                m = re.search(
                    r"DATE=\s*(\d?\d)[.\-/](\d\d)[.\-/](\d\d)", section_header, re.I
                )
                # assert isinstance(m, re.Match)  # TEMP for PLSX/PRSX reading
                if isinstance(m, re.Match):
                    short_year = int(m.group(3))
                    parsed_year = short_year + (
                        1900 if short_year >= now.year % 100 else 2000
                    )
                    new_date = datetime.date(
                        parsed_year, int(m.group(2)), int(m.group(1))
                    )
                else:
                    new_date = datetime.datetime.now().date()
                if new_date > date:
                    date = new_date
                data[i, ...] = np.fromfile(f, float, nd * nt, " ").reshape((nt, nd))
        attrs = {
            "date": date,
            "element": element_name,
            "year": year,
            "filename": filename,
        }
        _adf11 = DataArray(
            10 ** (data - 6),
            coords=[
                ("ion_charge", np.arange(zmin, zmax + 1, dtype=int)),
                ("electron_temperature", 10 ** (temperatures)),
                ("electron_density", 10 ** (densities + 6)),
            ],
            attrs=attrs,
        )
        assign_datatype(_adf11, quantity)
        return _adf11

    def get_adf15(
        self,
        element: str,
        charge: str,
        filetype: str,
        year="",
    ) -> DataArray:
        """Read data from the specified ADF15 ADAS file.

        Implementation is capable of reading files with compact and expanded formatting
        e.g. pec96][ne_pju][ne9.dat and pec40][ar_cl][ar16.dat respectively

        Parameters
        ----------
        element
            The atomic symbol for the element which will be retrieved.
        charge
            Charge state of the ion (e.g. 16 for Ar 16+), can also include
            other string for more complicated path (transport_llu][ar15ic.dat
            setting charge to "15ic")
        filetype
            The type of data to retrieve. Options: ic, cl, ca, ls, llu, ...
        year
            The two-digit year label for the data. = "transport" if special
            transport path


        Returns
        -------
        :
            The data in the specified file. Dimensions are density and
            temperature. Each members of the dataset correspond to a
            different charge state.

        """

        def explicit_reshape(data_to_reshape, nd, nt):
            data = np.empty((nd, nt))
            for id in range(nd):
                for it in range(nt):
                    data[id, it] = data_to_reshape[id * nt + it]

            return data

        def build_file_component(year, element):
            file_component = "transport"
            if year != "transport":
                file_component = f"pec{year}][{element.lower()}"

            return file_component

        def file_type(identifier):
            identifier_dict = {
                "+": "compact",
                ":": "expanded",
            }
            file_type = identifier_dict.get(identifier)
            if file_type is None:
                raise ValueError(f"Unknown file header identified ({identifier}).")

            return file_type

        def transition_match(transition_line):
            transition_type = "orbitals"
            match = (
                r"c\s+(\d+.)"  # isel
                r"\s+(\d+.\d+)"  # wavelength
                r"\s+(\d+)(\(\d\)\d\(.+\d?.\d\))-"  # transition upper level
                r".+(\d+)(\(\d\)\d\(.+\d?.\d\))"  # transition lower level
            )
            header_re = re.compile(match)
            m = header_re.search(transition_line)
            if not m:
                transition_type = "n_levels"
                match = r"c\s+(\d+.)\s+(\d+.\d+)\s+([n]\=.\d+.-.[n]\=.\d+)"
                header_re = re.compile(match)
                m = header_re.search(transition_line)
                if not m:
                    raise ValueError(f"Unknown transition formatting ({identifier}).")

            return transition_type, match

        file_component = build_file_component(year, element)
        filename = Path(pathname2url(file_component)) / pathname2url(
            f"{file_component}_{filetype.lower()}]"
            f"[{element.lower()}{charge.lower()}.dat"
        )

        header_match = {
            "compact": r"(\d+).+/(\S+).*\+(.*)photon",
            "expanded": r"(\d+).+/(\S+).*\:(.*)photon",
        }
        section_header_match = {
            "compact": r"(\d+.\d+).+\s+(\d+)\s+(\d+).+type\s?"
            r"=\s?(\S+).+isel.+\s+(\d+)",
            "expanded": r"(\d+.\d+)\s+(\d+)\s+(\d+).+type\s?="
            r"\s?(\S+).+isel\s+?=\s+?(\d+)",
        }
        with self._get_file("adf15", filename) as f:
            header = f.readline().strip().lower()
            identifier = file_type(header.split("/")[1][2])

            match = header_match[identifier]
            m = re.search(match, header, re.I)
            assert isinstance(m, re.Match)
            ntrans = int(m.group(1))
            element_name = m.group(2).strip().lower()
            charge_state = int(m.group(3))
            assert element_name == element.lower()
            m = re.search(r"(\d+)(\S*)", charge)
            assert isinstance(m, re.Match)
            extracted_charge = m.group(1)
            if charge_state != int(extracted_charge):
                raise ValueError(
                    f"Charge state in ADF15 file ({charge_state}) does not "
                    f"match argument ({charge})."
                )

            # Read first section header to build arrays outside of reading loop
            match = section_header_match[identifier]
            header_re = re.compile(match)
            m = None
            while not m:
                line = f.readline().strip().lower()
                m = header_re.search(line)
            assert isinstance(m, re.Match)
            nd = int(m.group(2))
            nt = int(m.group(3))
            ttype: List[str] = []
            tindex = np.empty(ntrans)
            wavelength = np.empty(ntrans)

            # Read Photon Emissivity Coefficient rates
            data = np.empty((ntrans, nd, nt))
            for i in range(ntrans):
                m = header_re.search(line)
                assert isinstance(m, re.Match)
                assert int(m.group(5)) - 1 == i
                tindex[i] = i + 1
                ttype.append(m.group(4))
                wavelength[i] = float(m.group(1))  # (Angstroms)

                densities = np.fromfile(f, float, nd, " ")
                temperatures = np.fromfile(f, float, nt, " ")
                data_tmp = np.fromfile(f, float, nd * nt, " ")
                data[i, :, :] = explicit_reshape(data_tmp, nd, nt)
                line = f.readline().strip().lower()

            data = np.transpose(np.array(data), (0, 2, 1))

            # Read Transition information from end of file
            file_end_re = re.compile(r"c\s+[isel].+\s+[transition].+\s+[type]")
            while not file_end_re.search(line):
                line = f.readline().strip().lower()
            _ = f.readline()
            if identifier == "expanded":
                _ = f.readline()
            line = f.readline().strip().lower()
            transition_type, match = transition_match(line)
            transition_re = re.compile(match)

            format_transition = {
                "orbitals": lambda m: f"{m.group(4)}-{m.group(6)}".replace(" ", ""),
                "n_levels": lambda m: m.group(3).replace(" ", ""),
            }
            transition = []
            for i in tindex:
                m = transition_re.search(line)
                assert isinstance(m, re.Match)
                assert int(m.group(1)[:-1]) == i
                transition_tmp = format_transition[transition_type](m)
                transition.append(transition_tmp)
                line = f.readline().strip().lower()
        attrs = {
            "filename": filename,
        }

        coords = [
            ("index", tindex),
            ("electron_temperature", temperatures),  # eV
            ("electron_density", densities * 10**6),  # m**-3
        ]

        pecs = DataArray(
            data * 10**-6,
            coords=coords,
            attrs=attrs,
        )

        # Add extra dimensions attached to index
        pecs = pecs.assign_coords(wavelength=("index", wavelength))  # (A)
        pecs = pecs.assign_coords(
            transition=("index", transition)
        )  # (2S+1)L(w-1/2)-(2S+1)L(w-1/2) of upper-lower levels, no blank spaces
        pecs = pecs.assign_coords(type=("index", ttype))  # (excit, recomb, cx)

        assign_datatype(pecs, "pec")
        return pecs

    def _get_file(self, dataclass: str, filename: Union[str, Path]) -> TextIO:
        """Retrieves an ADAS file, downloading it from OpenADAS if
        necessary. It will cache any downloads for later use.

        Parameters
        ----------
        dataclass
            The format of ADAS data in this file (e.g., ADF11).
        filename
            Name of the file to get.

        Returns
        -------
        :
            A file-like object from which the data can be read.

        """
        filepath = self.path / dataclass / filename
        if self.openadas and not filepath.exists():
            filepath.parent.mkdir(parents=True, exist_ok=True)
            urlretrieve(
                f"https://open.adas.ac.uk/download/{dataclass}/{filename}", filepath
            )
        return filepath.open("r")

    @property
    def requires_authentication(self) -> Literal[False]:
        """Reading ADAS data never requires authentication."""
        return False



import copy
from typing import cast
from typing import List
from typing import Tuple

import matplotlib.pylab as plt
import numpy as np
from numpy.core.numeric import zeros_like
from pandas import DataFrame
import scipy
import xarray as xr
from xarray import DataArray

from indica.numpy_typing import LabeledArray
from indica.profilers.profiler_gauss import ProfilerGauss
from indica.readers.adas import ADASReader
from indica.readers.adas import ADF11
from indica.utilities import DATA_PATH
from indica.utilities import set_plot_colors
from .abstractoperator import Operator

np.set_printoptions(edgeitems=10, linewidth=100)


class FractionalAbundance(Operator):
    """Calculate fractional abundance for all ionisation charges of a given element.

    scd
        effective ionisation rate coefficients
    acd
        effective recombination rate coefficients
    ccd
        charge exchange recombination coefficients
    """

    def __init__(
        self,
        scd: DataArray,
        acd: DataArray,
        ccd: DataArray = None,
    ):
        self.Ne = None
        self.Te = None
        self.Nh = None
        self.tau = None
        self.F_z_t0 = None
        self.scd = scd
        self.acd = acd
        self.ccd = ccd

    def interpolate_rates(
        self,
        Ne: DataArray,
        Te: DataArray,
    ):
        """Interpolates rates based on inputted Ne and Te, also determines the number
        of ionisation charges for a given element.

        Ne
            electron density profile
        Te
            electron temperature profile
        """

        self.Ne, self.Te = Ne, Te  # type: ignore

        scd_spec = self.scd.indica.interp2d(
            electron_temperature=Te,
            electron_density=Ne,
            method="cubic",
            assume_sorted=True,
        )

        if self.ccd is not None:
            ccd_spec = self.ccd.indica.interp2d(
                electron_temperature=Te,
                electron_density=Ne,
                method="cubic",
                assume_sorted=True,
            )
        else:
            ccd_spec = None

        acd_spec = self.acd.indica.interp2d(
            electron_temperature=Te,
            electron_density=Ne,
            method="cubic",
            assume_sorted=True,
        )

        self.scd_spec, self.acd_spec, self.ccd_spec = scd_spec, acd_spec, ccd_spec
        self.num_of_ion_charge = self.scd_spec.shape[0] + 1

        return scd_spec, acd_spec, ccd_spec, self.num_of_ion_charge

    def calc_ionisation_balance_matrix(
        self,
        Ne: DataArray,
        Nh: DataArray = None,
    ):
        """Calculates the ionisation balance matrix
        Ne
            electron density profile
        Nh
            thermal neutral hydrogen profile
        """
        if Nh is not None:
            if self.ccd is None:
                raise ValueError(
                    "Nh (Thermal hydrogen density) cannot be given when \
                    ccd (effective charge exchange recombination) at initialisation \
                    is None."
                )
        elif self.ccd is not None:
            Nh = cast(DataArray, zeros_like(Ne))

        self.Ne, self.Nh = Ne, Nh  # type: ignore

        num_of_ion_charge = self.num_of_ion_charge
        scd, acd, ccd = self.scd_spec, self.acd_spec, self.ccd_spec

        x1_coord = scd.coords[[k for k in scd.dims if k != "ion_charge"][0]]
        self.x1_coord = x1_coord

        dims = (
            num_of_ion_charge,
            num_of_ion_charge,
            *x1_coord.shape,
        )

        ionisation_balance_matrix = np.zeros(dims)

        icharge = 0
        ionisation_balance_matrix[icharge, icharge : icharge + 2] = np.array(
            [
                -Ne * scd[icharge],  # type: ignore
                Ne * acd[icharge]
                + (Nh * ccd[icharge] if Nh is not None and ccd is not None else 0.0),
            ]
        )
        for icharge in range(1, num_of_ion_charge - 1):
            ionisation_balance_matrix[icharge, icharge - 1 : icharge + 2] = np.array(
                [
                    Ne * scd[icharge - 1],
                    -Ne * (scd[icharge] + acd[icharge - 1])  # type: ignore
                    - (
                        Nh * ccd[icharge - 1]
                        if Nh is not None and ccd is not None
                        else 0.0
                    ),
                    Ne * acd[icharge]
                    + (
                        Nh * ccd[icharge] if Nh is not None and ccd is not None else 0.0
                    ),
                ]
            )
        icharge = num_of_ion_charge - 1
        ionisation_balance_matrix[icharge, icharge - 1 : icharge + 1] = np.array(
            [
                Ne * scd[icharge - 1],
                -Ne * (acd[icharge - 1])  # type: ignore
                - (
                    Nh * ccd[icharge - 1] if Nh is not None and ccd is not None else 0.0
                ),
            ]
        )

        ionisation_balance_matrix = np.squeeze(ionisation_balance_matrix)
        self.ionisation_balance_matrix = ionisation_balance_matrix

        return ionisation_balance_matrix

    def calc_F_z_tinf(
        self,
    ):
        """Calculates the equilibrium fractional abundance of all ionisation charges,
        F_z(t=infinity) used for the final time evolution equation.
        """
        x1_coord = self.x1_coord
        ionisation_balance_matrix = self.ionisation_balance_matrix

        null_space = np.zeros((self.num_of_ion_charge, x1_coord.size))
        F_z_tinf = np.zeros((self.num_of_ion_charge, x1_coord.size))

        for ix1 in range(x1_coord.size):
            null_space[:, ix1, np.newaxis] = scipy.linalg.null_space(
                ionisation_balance_matrix[:, :, ix1]
            )

        # Complex type casting for compatibility with eigen calculation results later.
        F_z_tinf = np.abs(null_space).astype(dtype=np.complex128)

        # normalization needed for high-z elements
        F_z_tinf = F_z_tinf / np.sum(F_z_tinf, axis=0)

        F_z_tinf = DataArray(
            data=F_z_tinf,
            coords=[
                (
                    "ion_charge",
                    np.linspace(0, self.num_of_ion_charge - 1, self.num_of_ion_charge),
                ),
                x1_coord,
            ],
            dims=["ion_charge", x1_coord.dims[0]],
        )

        self.F_z_tinf = F_z_tinf

        return np.real(F_z_tinf)

    def calc_eigen_vals_and_vecs(
        self,
    ):
        """Calculates the eigenvalues and eigenvectors of the ionisation balance
        matrix.
        """
        x1_coord = self.x1_coord
        eig_vals = np.zeros(
            (self.num_of_ion_charge, x1_coord.size), dtype=np.complex128
        )
        eig_vecs = np.zeros(
            (self.num_of_ion_charge, self.num_of_ion_charge, x1_coord.size),
            dtype=np.complex128,
        )

        for ix1 in range(x1_coord.size):
            eig_vals[:, ix1], eig_vecs[:, :, ix1] = scipy.linalg.eig(
                self.ionisation_balance_matrix[:, :, ix1],
            )

        self.eig_vals = eig_vals
        self.eig_vecs = eig_vecs

        return eig_vals, eig_vecs

    def calc_eigen_coeffs(
        self,
        F_z_t0: DataArray = None,
    ):
        """Calculates the coefficients from the eigenvalues and eigenvectors for the
        time evolution equation.

        F_z_t0
            Initial fractional abundance for given impurity element. (Optional)
        """
        x1_coord = self.x1_coord

        if F_z_t0 is None:
            # mypy doesn't understand contionals or reassignments either.
            F_z_t0 = np.zeros(self.F_z_tinf.shape, dtype=np.complex128)  # type: ignore
            F_z_t0[0, :] = np.array(  # type: ignore
                [1.0 + 0.0j for i in range(x1_coord.size)]
            )

            F_z_t0 = DataArray(
                data=F_z_t0,
                coords=[
                    (
                        "ion_charge",
                        np.linspace(
                            0, self.num_of_ion_charge - 1, self.num_of_ion_charge
                        ),
                    ),
                    x1_coord,  # type: ignore
                ],
                dims=["ion_charge", x1_coord.dims[0]],
            )
        else:
            try:
                assert F_z_t0.ndim < 3
            except AssertionError:
                raise ValueError("F_z_t0 must be at most 2-dimensional.")

            F_z_t0 = F_z_t0 / np.sum(F_z_t0, axis=0)
            F_z_t0 = F_z_t0.as_type(dtype=np.complex128)  # type: ignore

            F_z_t0 = DataArray(
                data=F_z_t0.values,  # type: ignore
                coords=[
                    (
                        "ion_charge",
                        np.linspace(
                            0, self.num_of_ion_charge - 1, self.num_of_ion_charge
                        ),
                    ),
                    x1_coord,  # type: ignore
                ],
                dims=["ion_charge", x1_coord.dims[0]],
            )

        eig_vals = self.eig_vals
        eig_vecs_inv = np.zeros(self.eig_vecs.shape, dtype=np.complex128)
        for ix1 in range(x1_coord.size):
            eig_vecs_inv[:, :, ix1] = np.linalg.pinv(
                np.transpose(self.eig_vecs[:, :, ix1])
            )

        boundary_conds = F_z_t0 - self.F_z_tinf

        eig_coeffs = np.zeros(eig_vals.shape, dtype=np.complex128)
        for ix1 in range(x1_coord.size):
            eig_coeffs[:, ix1] = np.dot(boundary_conds[:, ix1], eig_vecs_inv[:, :, ix1])

        self.eig_coeffs = eig_coeffs

        # If argument to numpy functions is of type DataArray then output is of
        # type DataArray
        F_z_t0 = np.abs(np.real(F_z_t0))  # type: ignore

        self.F_z_t0 = F_z_t0  # type: ignore

        return eig_coeffs, F_z_t0

    def calculate_abundance(self, tau: LabeledArray):
        """Calculates the fractional abundance of all ionisation charges at time tau.

        tau
            Time after t0 (t0 is defined as the time at which F_z_t0 is taken).
        """
        x1_coord = self.x1_coord
        F_z_t = copy.deepcopy(self.F_z_tinf)
        for ix1 in range(x1_coord.size):
            if isinstance(tau, (DataArray, np.ndarray)):
                itau = tau[ix1].values if isinstance(tau, DataArray) else tau[ix1]
            else:
                itau = tau

            for icharge in range(self.num_of_ion_charge):
                F_z_t[:, ix1] += (
                    self.eig_coeffs[icharge, ix1]
                    * np.exp(self.eig_vals[icharge, ix1] * itau)
                    * self.eig_vecs[:, icharge, ix1]
                )

        F_z_t = np.abs(np.real(F_z_t))

        self.F_z_t = F_z_t
        # Mypy complains about assigning a LabeledArray to an object that was type None.
        # Can't really fix incompatibility without eliminating LabeledArray since mypy
        # seems to have an issue with type aliases.
        self.tau = tau  # type: ignore

        return F_z_t

    def __call__(  # type: ignore
        self,
        Te: DataArray,
        Ne: DataArray = None,
        Nh: DataArray = None,
        tau: LabeledArray = None,
        F_z_t0: DataArray = None,
        full_run: bool = False,
    ) -> DataArray:
        """Executes all functions in correct order to calculate the fractional
        abundance.

        Ne
            electron density profile
        Te
            electron temperature profile
        Nh
            thermal neutral hydrogen profile
        tau
            Time after t0 (t0 defined as the time at which F_z_t0 is calculated)
        F_z_t0
            Initial fractional abundance at t0
        full_run
            Boolean specifying whether to run the entire ordered workflow
        """
        if full_run or not hasattr(self, "F_z_t"):
            self.interpolate_rates(Ne, Te)

            self.calc_ionisation_balance_matrix(Ne, Nh)

            self.calc_F_z_tinf()

            if tau is None:
                F_z_t = np.real(self.F_z_tinf)
                self.F_z_t = F_z_t
                return F_z_t

            self.calc_eigen_vals_and_vecs()

            self.calc_eigen_coeffs(F_z_t0)

            F_z_t = self.calculate_abundance(tau)

            self.F_z_t = F_z_t
        else:

            F_z_t = interpolate_results(self.F_z_t, self.Te, Te)

        return F_z_t


class PowerLoss(Operator):
    """Calculate the total power loss associated with a given impurity element

    plt
        line emission emissivity coefficients
    prb
        recombination and bremsstrahlung emissivity coefficients
    prc
        charge exchange emissivity coefficients
    """

    def __init__(
        self,
        plt: DataArray,
        prb: DataArray,
        prc: DataArray = None,
    ):
        self.plt = plt
        self.prc = prc
        self.prb = prb
        self.Ne = None
        self.Nh = None
        self.Te = None
        self.F_z_t = None

        imported_data = {}
        imported_data["plt"] = self.plt
        imported_data["prb"] = self.prb
        if self.prc is not None:
            imported_data["prc"] = self.prc

    def interpolate_power(
        self,
        Ne: DataArray,
        Te: DataArray,
    ):
        """Interpolates the various powers based on inputted Ne and Te.

        Ne
            electron density profile
        Te
            electron temperature profile
        """

        self.Ne, self.Te = Ne, Te  # type: ignore
        plt_spec = self.plt.interp(electron_temperature=Te, method="cubic").interp(
            electron_density=Ne, method="linear"
        )

        if self.prc is not None:
            prc_spec = self.prc.interp(electron_temperature=Te, method="cubic").interp(
                electron_density=Ne, method="linear"
            )

        else:
            prc_spec = None

        prb_spec = self.prb.interp(electron_temperature=Te, method="cubic").interp(
            electron_density=Ne, method="linear"
        )

        self.plt_spec, self.prc_spec, self.prb_spec = plt_spec, prc_spec, prb_spec
        self.num_of_ion_charge = self.plt_spec.shape[0] + 1

        return plt_spec, prc_spec, prb_spec, self.num_of_ion_charge

    def calculate_power_loss(
        self,
        Ne: DataArray,
        F_z_t: DataArray,
        Nh: DataArray = None,
    ):
        """Calculates total radiated power of all ionisation charges of a given
        impurity element.

        Ne
            electron density profile
        F_z_t
            fractional abundance of all ionisation charges of given element.
        Nh
            thermal neutral hydrogen density profile

        """
        if Nh is not None:
            if self.prc is None:
                raise ValueError(
                    "Nh (Thermal hydrogen density) cannot be given when \
                    prc (effective charge exchange power) at initialisation \
                    is None."
                )
        elif self.prc is not None:
            Nh = cast(DataArray, zeros_like(Ne))

        self.Ne, self.Nh = Ne, Nh  # type: ignore

        if F_z_t is not None:
            try:
                assert not np.iscomplexobj(F_z_t)
            except AssertionError:
                raise ValueError(
                    "Inputted F_z_t is a complex type or array of complex numbers, \
                        must be real"
                )
            self.F_z_t = F_z_t  # type: ignore
        elif self.F_z_t is None:
            raise ValueError("Please provide a valid F_z_t (Fractional Abundance).")

        self.x1_coord = self.plt_spec.coords[
            [k for k in self.plt_spec.dims if k != "ion_charge"][0]
        ]

        x1_coord = self.x1_coord

        plt, prb, prc = self.plt_spec, self.prb_spec, self.prc_spec

        # TODO: make this faster by using DataArray methods
        cooling_factor = xr.zeros_like(self.F_z_t)
        for ix1 in range(x1_coord.size):
            icharge = 0
            cooling_factor[icharge, ix1] = (
                plt[icharge, ix1] * self.F_z_t[icharge, ix1]  # type: ignore
            )
            for icharge in range(1, self.num_of_ion_charge - 1):
                cooling_factor[icharge, ix1] = (
                    plt[icharge, ix1]
                    + (
                        (Nh[ix1] / Ne[ix1]) * prc[icharge - 1, ix1]
                        if (prc is not None) and (Nh is not None)
                        else 0.0
                    )
                    + prb[icharge - 1, ix1]
                ) * self.F_z_t[
                    icharge, ix1
                ]  # type: ignore

            icharge = self.num_of_ion_charge - 1
            cooling_factor[icharge, ix1] = (
                (
                    (Nh[ix1] / Ne[ix1]) * prc[icharge - 1, ix1]
                    if (prc is not None) and (Nh is not None)
                    else 0.0
                )
                + prb[icharge - 1, ix1]
            ) * self.F_z_t[
                icharge, ix1
            ]  # type: ignore

        self.cooling_factor = cooling_factor

        return cooling_factor

    def __call__(  # type: ignore
        self,
        Te: DataArray,
        F_z_t: DataArray,
        Ne: DataArray = None,
        Nh: DataArray = None,
        full_run: bool = False,
    ):
        """Executes all functions in correct order to calculate the total radiated
        power.

        Ne
            electron density profile
        Te
            electron temperature profile
        Nh
            thermal neutral hydrogen density profile
        F_z_t
            fractional abundance of all ionisation charges of given element
        full_run
            Boolean specifying whether to run the entire ordered workflow
        """

        if full_run or not hasattr(self, "cooling_factor"):
            self.interpolate_power(Ne, Te)
            cooling_factor = self.calculate_power_loss(Ne, F_z_t, Nh)  # type: ignore
            self.cooling_factor = cooling_factor
        else:
            cooling_factor = interpolate_results(self.cooling_factor, self.Te, Te)

        return cooling_factor


def interpolate_results(
    data: DataArray, Te_data: DataArray, Te_interp: DataArray, method="cubic"
):
    """
    Interpolate fractional abundance or cooling factor on electron
    temperature for fast processing

    atomic_data
        Fractional abundance or cooling factor DataArrays
    Te
        Electron temperature on which interpolation is to be performed
    """
    dim_old = [d for d in data.dims if d != "ion_charge"][0]
    _data = data.assign_coords(electron_temperature=(dim_old, Te_data.data))
    _data = _data.swap_dims({dim_old: "electron_temperature"}).drop_vars(dim_old)
    result = _data.interp(electron_temperature=Te_interp).drop_vars(
        ("electron_temperature",)
    )
    return result


def default_atomic_data(
    elements: Tuple[str, ...],
    Te: DataArray = None,
    Ne: DataArray = None,
    Nh: DataArray = None,
    tau: DataArray = None,
):
    """
    Initialises atomic data classes with default ADAS files and runs the
    __call__ with default plasma parameters
    """
    if Te is None or Ne is None:
        Te, Ne, Nh, tau = default_profiles()

    # print_like("Initialize fractional abundance and power loss objects")
    # fract_abu, power_loss_tot, power_loss_sxr = {}, {}, {}
    fract_abu, power_loss_tot = {}, {}
    adas_reader = ADASReader()
    for elem in elements:
        scd = adas_reader.get_adf11("scd", elem, ADF11[elem]["scd"])
        acd = adas_reader.get_adf11("acd", elem, ADF11[elem]["acd"])
        ccd = adas_reader.get_adf11("ccd", elem, ADF11[elem]["ccd"])
        fract_abu[elem] = FractionalAbundance(scd, acd, ccd=ccd)
        F_z_t = fract_abu[elem](Ne=Ne, Te=Te, Nh=Nh, tau=tau)
        plt = adas_reader.get_adf11("plt", elem, ADF11[elem]["plt"])
        prb = adas_reader.get_adf11("prb", elem, ADF11[elem]["prb"])
        prc = adas_reader.get_adf11("prc", elem, ADF11[elem]["prc"])
        power_loss_tot[elem] = PowerLoss(plt, prb, prc=prc)
        power_loss_tot[elem](Te, F_z_t, Ne=Ne, Nh=Nh)

        # try:
        #     pls = adas_reader.get_adf11("pls", elem, ADF11[elem]["pls"])
        #     prs = adas_reader.get_adf11("prs", elem, ADF11[elem]["prs"])
        #     power_loss_sxr[elem] = PowerLoss(pls, prs)
        #     power_loss_sxr[elem](Te, F_z_t, Ne=Ne, Nh=Nh)
        # except Exception:
        #     print(f"No SXR-filtered data available for element {elem}")

    return fract_abu, power_loss_tot  # , power_loss_sxr


def default_profiles(n_rad: int = 20):
    """
    Set default plasma profiles to calculate atomic data
    """
    xend = 1.02
    rho_end = 1.01
    rho = np.abs(np.linspace(rho_end, 0, n_rad) ** 1.8 - rho_end - 0.01)
    rho_coord = xr.DataArray(rho, coords={"rhop": rho}, dims="rhop").coords
    Te = xr.DataArray(np.linspace(50, 10e3, n_rad), coords=rho_coord)
    Ne = xr.DataArray(np.logspace(18, 21, n_rad), coords=rho_coord)

    # TODO: fix FractionalAbundance so that it does 2d interp of Nh and Te
    params = {
        "y0": 1e14,
        "y1": 5e15,
        "yend": 5e15,
        "wcenter": 0.01,
        "wped": 18,
        "peaking": 1,
    }
    Nh_prof = ProfilerGauss(
        datatype="neutral_density", xspl=rho, xend=xend, parameters=params
    )
    Nh = Nh_prof()
    tau = None
    return Te, Ne, Nh, tau


def cooling_factor_corona(
    elements: List[str],
    write_to_file: bool = False,
    plot: bool = False,
    new_figure: bool = True,
    include_neutrals: bool = False,
):
    """
    Initialises atomic data classes with default ADAS files and runs the
    __call__ with default plasma parameters
    """
    tau = None
    Ne_const = 5.0e19
    Nh1 = 1.0e17
    Nh0 = 1.0e12

    fract_abu: dict = {}
    power_loss_tot: dict = {}
    atomic_data_files: dict = {}
    cooling_factor: dict = {}
    filenames = ""
    files_to_read = ["scd", "acd", "ccd", "plt", "prb", "prc"]
    Te_files = []

    print("Read atomic data")
    adas_reader = ADASReader()
    for elem in elements:
        atomic_data_files[elem] = {}
        for file_type in files_to_read:
            _atomic_data = adas_reader.get_adf11(
                file_type, elem, ADF11[elem][file_type]
            )
            filenames += f"{_atomic_data.filename}"
            Te_files.append(_atomic_data.electron_temperature)
            atomic_data_files[elem][file_type] = _atomic_data

    # Set Te so that max(Te) doesn't exceed the value available in all atomic-data files
    _indx = np.argmin(np.array([np.max(_Te) for _Te in Te_files]))
    _Te = Te_files[_indx]
    nTe = np.size(_Te)
    Te = DataArray(_Te.data, coords=[("index", np.arange(nTe))])
    Ne = xr.full_like(Te, Ne_const)
    Nh = xr.full_like(Te, 0.0)
    if include_neutrals:
        _Nh = np.array([Te.values[i] for i in np.arange(Te.size - 1, -1, -1)])
        _Nh -= np.min(_Nh)
        _Nh /= np.max(_Nh)
        _Nh *= Nh1
        _Nh += Nh0
        Nh.values = _Nh

    _to_write = {"Te": np.array(Te), "Ne": np.array(Ne), "Nh": np.array(Nh)}

    print("Calculate fractional abundance and cooling factors")
    for elem in elements:
        print(f"  {elem}")
        fract_abu[elem] = FractionalAbundance(
            atomic_data_files[elem]["scd"],
            atomic_data_files[elem]["acd"],
            ccd=atomic_data_files[elem]["ccd"],
        )
        _fz = fract_abu[elem](Ne=Ne, Te=Te, Nh=Nh, tau=tau)

        power_loss_tot[elem] = PowerLoss(
            atomic_data_files[elem]["plt"],
            atomic_data_files[elem]["prb"],
            prc=atomic_data_files[elem]["prc"],
        )
        _power_loss = power_loss_tot[elem](Te, _fz, Ne=Ne, Nh=Nh)

        _cooling_factor: DataArray = _power_loss.sum("ion_charge")
        _cooling_factor = (
            _cooling_factor.assign_coords(electron_temperature=("index", Te.data))
            .swap_dims({"index": "electron_temperature"})
            .drop_vars("index")
        )

        cooling_factor[elem] = _cooling_factor
        _to_write[elem] = np.array(_cooling_factor)

    _to_write["atomic_data_files"] = filenames

    if write_to_file:
        if include_neutrals:
            file_name = f"{DATA_PATH}corona_cooling_factors_Nh.csv"
        else:
            file_name = f"{DATA_PATH}corona_cooling_factors.csv"
        print(f"Writing data to {file_name}")
        df = DataFrame(_to_write)
        df.to_csv(file_name)

    if plot:
        if new_figure:
            plt.figure()
        cmap, _ = set_plot_colors()
        cols = cmap(np.linspace(0.75, 0.1, len(cooling_factor), dtype=float))

        label = ""
        marker = "o"
        linestyle = "solid"
        if include_neutrals:
            marker = ""
            linestyle = "dashed"

        for i, elem in enumerate(elements):
            if new_figure:
                label = elem
            cooling_factor[elem].plot(
                label=label,
                alpha=0.8,
                marker=marker,
                color=cols[i],
                linestyle=linestyle,
            )
        if new_figure:
            plt.xscale("log")
            plt.yscale("log")
            plt.legend()

    return cooling_factor, _to_write, fract_abu
